import os
import warnings
import typing
from typing import Any, Optional
import logging

import ncempy.io.dm
import numpy as np

from libertem.common.math import prod
from libertem.common import Shape
from .base import BasePartition, DataSetException, DataSetMeta, File, IOBackend, DataSet
from .dm import DMDataSet, SingleDMDatasetParams, DMFileSet

log = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from numpy import typing as nt
    from libertem.common.executor import JobExecutor


class SingleDMDataSet(DMDataSet):
    """
    Reader for a single DM3/DM4 file. Handles 4D-STEM, 3D-Spectrum Images,
    and TEM image stacks stored in a single-file format. Where possible
    the structure will be inferred from the file metadata.

    .. versionadded:: 0.11.0

    Note
    ----
    Single-file DM data can be stored on disk using either normal C-ordering,
    which is an option in recent versions of GMS, or an alternative F/C-hybrid
    ordering depending on the imaging mode and dimensionality. The reading of
    F/C-hybrid files is currently not supported for performance reasons.

    The DataSet will try to infer the ordering from the file metadata and
    read accordingly. If the file uses the older hybrid F/C-ordering
    :code:`(flat_sig, flat_nav)` then the dataset will raise an exception
    unless the `force_c_order` argument. is set to true.

    A converter for F/C-hybrid files is provided as
    :meth:`~libertem.contrib.convert_transposed.convert_dm4_transposed`.

    Note
    ----
    In the Web-GUI a 2D-image or 3D-stack/spectrum image will have extra
    singleton navigation dimensions prepended to allow them to display.
    DM files containing multiple datasets are supported via the
    `dataset_index` argument.

    While capable of reading 2D/3D files, LiberTEM is not particularly
    well-adapted to processing these data and the user should consider
    other tools. Individual spectra or vectors (1D data) are not supported.

    Parameters
    ----------

    path : PathLike
        The path to the .dm3/.dm4 file

    nav_shape : Tuple[int, ...], optional
        Over-ride the nav_shape provided by the file metadata.
        This can be used to adjust the total
        number of frames.

    sig_shape: Tuple[int, ...], optional
        Over-ride the sig_shape provided by the file metadata.
        Data are read sequentially in all cases, therefore this
        is typically only interesting if the total number of
        sig pixels remains constant.

    sync_offset: int, optional, by default 0
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    io_backend: IOBackend, optional
        A specific IOBackend implementation to over-ride the
        platform default.

    force_c_order: bool, optional, by default False
        Force the data to be interpreted as a C-ordered
        array regardless of the tag information. This will lead
        to incorrect results on an hybrid C/F-ordered file.

    dataset_index: int, optional
        In the case of a multi-dataset DM file this can
        be used to open a specific dataset index. Note that
        the datasets in a DM-file often begin with a thumbnail
        which occupies the 0 dataset index. If not provided the
        first compatible dataset found in the file is used.
    """
    def __init__(
        self,
        path: os.PathLike,
        nav_shape: Optional[tuple[int, ...]] = None,
        sig_shape: Optional[tuple[int, ...]] = None,
        sync_offset: int = 0,
        io_backend: Optional[IOBackend] = None,
        force_c_order: bool = False,
        dataset_index: Optional[int] = None
    ):
        super().__init__(io_backend=io_backend)
        self._filesize = None

        self._path = path
        self._nav_shape = tuple(nav_shape) if nav_shape else None
        self._sig_shape = tuple(sig_shape) if sig_shape else None
        self._sync_offset = sync_offset
        self._force_c_order = force_c_order
        self._dm_ds_index = dataset_index

    def __new__(cls, *args, **kwargs):
        '''
        Skip the superclasse's :code:`__new__()` method.

        Instead, go straight to the grandparent. That disables the
        :class:`DMDataSet` type determination magic. Otherwise unpickling will
        always yield a :class:`SingleDMDataSet` since this class inherits the
        parent's :code:`__new__()` method and unpickling calls it without
        parameters, making it select :class:`SingleDMDataSet`.

        It mimics calling the superclass :code:`__new__(cls)` without additional
        parameters, just like the parent's method.
        '''
        return DataSet.__new__(cls)

    def __repr__(self):
        try:
            shape = f' - {self.shape}'
        except AttributeError:
            shape = ''
        return f"<DMFileDataset {self._path}>" + shape

    @property
    def dtype(self) -> "nt.DTypeLike":
        return self.meta.raw_dtype

    @property
    def shape(self):
        return self.meta.shape

    @classmethod
    def get_supported_extensions(cls):
        return {"dm3", "dm4"}

    def _get_filesize(self):
        return os.stat(self._path).st_size

    @classmethod
    def get_msg_converter(cls):
        return SingleDMDatasetParams

    @classmethod
    def detect_params(cls, path: str, executor):
        pathlow = path.lower()
        if pathlow.endswith(".dm3") or pathlow.endswith(".dm4"):
            array_meta = executor.run_function(cls._read_metadata, path)
            sig_dims = array_meta['sig_dims']
            if sig_dims == 1:
                return False
            sync_offset = 0
            nav_shape, sig_shape = cls._modify_shape(array_meta['shape'],
                                                    sig_dims=sig_dims)
            if len(nav_shape) == 1:
                nav_shape = (1,) + nav_shape
            image_count = prod(nav_shape)
        else:
            return False
        return {
            "parameters": {
                "path": path,
                "nav_shape": nav_shape,
                "sig_shape": sig_shape,
                "sync_offset": sync_offset,
            },
            "info": {
                "image_count": image_count,
                "native_sig_shape": sig_shape,
            }
        }

    def check_valid(self):
        try:
            with ncempy.io.dm.fileDM(self._path, on_memory=True):
                pass
            return True
        except OSError as e:
            raise DataSetException("invalid dataset: %s" % e)

    @classmethod
    def _read_metadata(cls, path, use_ds=None):
        with ncempy.io.dm.fileDM(path, on_memory=True) as fp:
            tags = fp.allTags
            array_map = {}
            start_from = 1 if fp.thumbnail else 0
            for ds_idx in range(start_from, fp.numObjects):
                dims = fp.dataShape[ds_idx]
                if dims < 2:
                    # Spectrum-only ?
                    continue
                shape = (fp.xSize[ds_idx], fp.ySize[ds_idx])
                if dims > 2:
                    shape = shape + (fp.zSize[ds_idx],)
                if dims > 3:
                    shape = shape + (fp.zSize2[ds_idx],)
                array_map[ds_idx] = {'shape': shape, 'ds_idx': ds_idx}
                array_map[ds_idx]['offset'] = fp.dataOffset[ds_idx]
                try:
                    array_map[ds_idx]['dtype'] = fp._DM2NPDataType(fp.dataType[ds_idx])
                except OSError:
                    # unconvertible DM data type
                    array_map[ds_idx]['dtype'] = fp.dataType[ds_idx]

        if not array_map:
            raise DataSetException('Unable to find any 2/3/4D datasets in DM file')

        if use_ds is not None:
            if use_ds in array_map.keys():
                ds_idx = use_ds
            else:
                raise DataSetException(f'Specified dataset idx {use_ds} not found in file')
        else:
            # Use first dataset index we loaded
            ds_idx = [*array_map.keys()][0]
            if len(array_map) > 1:
                warnings.warn(
                    "Found multiple datasets in DM file, using first dataset",
                    RuntimeWarning
                )

        array_meta = array_map[ds_idx]
        ndims = len(array_meta['shape'])
        # Set default metadata in case tags are incomplete
        array_meta['format'] = 'Unknown'
        array_meta['sig_dims'] = 2
        # Assume C-ordering for 2D images and 3D image stacks
        # Assume F-ordering for STEM data unless tagged otherwise
        # Spectrum images are also F-ordered but these data must
        # be recognized from the tags (they can be 2- or 3-D)
        array_meta['c_order'] = True if ndims in (2, 3) else False

        # Infer array ordering
        nest = cls._tags_to_nest(tags)
        # Must + 1 because DM uses 1-based-indexing in its tags
        dm_data_key = str(array_meta['ds_idx'] + 1)
        try:
            data_tags = nest['ImageList'][dm_data_key]['ImageTags']
        except KeyError:
            # unrecognized / invalid tag structure, return defaults
            return array_meta

        if 'Meta Data' in data_tags:
            meta_data = data_tags['Meta Data']
            array_meta['format'] = meta_data.get('Format', 'Unknown')
            if str(array_meta['format']).strip().lower() == 'spectrum image':
                assert ndims in (2, 3)
                array_meta['sig_dims'] = 1
                if ndims == 3:
                    # 3-D spectrum images seem to be F-ordered
                    # 2-D SI are seemingly C-ordered (value set above)
                    array_meta['c_order'] = False
            if 'Data Order Swapped' in meta_data:
                # Always defer to tag for ordering if available
                # This line handes the new-style STEM datasets
                # The bool(int()) is just-in-case for string tags
                array_meta['c_order'] = bool(int(meta_data['Data Order Swapped']))

        # Need to find a 3D image stack with the 'Meta Data' + 'Format' tags
        if array_meta['format'] not in ('Spectrum image',
                                        'Image',
                                        'Diffraction image'):
            warnings.warn(
                f"Unrecognized image format {array_meta['format']}, "
                "DM tags may be parsed incorrectly",
                RuntimeWarning
            )
        return array_meta

    @staticmethod
    def _tags_to_nest(tags: dict[str, Any]):
        tags_nest = {}
        for tag, element in tags.items():
            tag = tag.strip('.')
            _insert_to = tags_nest
            for tag_el in tag.split('.')[:-1]:
                try:
                    _insert_to = _insert_to[tag_el]
                except KeyError:
                    _insert_to[tag_el] = {}
                    _insert_to = _insert_to[tag_el]
            _insert_to[tag.split('.')[-1]] = element
        return tags_nest

    @staticmethod
    def _modify_shape(shape: tuple[int, ...], sig_dims: int = 2):
        # The shape reversal to read in C-ordering applies to DM4/STEM files
        # saved in the new style as well as DM3, 3D image stacks saved
        # in older versions of GMS. Must check whether newer image stacks
        # are saved in C-ordering as well (despite the metadata order)
        shape = tuple(reversed(shape))
        shape = tuple(map(int, shape))
        nav_shape = shape[:-sig_dims]
        sig_shape = shape[-sig_dims:]
        if not nav_shape:
            # Special case for 2D image data, LT always requires a nav dim
            nav_shape = (1,)
        return nav_shape, sig_shape

    def initialize(self, executor: 'JobExecutor'):
        self._filesize = executor.run_function(self._get_filesize)
        array_meta = executor.run_function(self._read_metadata,
                                           self._path,
                                           use_ds=self._dm_ds_index)
        sig_dims = array_meta['sig_dims']
        self._array_offset = array_meta['offset']
        self._raw_dtype = array_meta['dtype']
        assert self._raw_dtype is not None and not isinstance(self._raw_dtype, int)
        array_c_ordered = self._force_c_order or array_meta['c_order']

        if not array_c_ordered:
            raise DataSetException('Cannot identify DM file as C-ordered from metadata'
                                   'use force_c_order=True to force behaviour.')

        nav_shape, sig_shape = self._modify_shape(array_meta['shape'],
                                                  sig_dims=sig_dims)

        # Image count is true number of frames in file (?)
        self._image_count = int(prod(nav_shape))
        if self._nav_shape is not None:
            manual_nav_shape_product = prod(self._nav_shape)
            if manual_nav_shape_product > self._image_count:
                raise DataSetException('Specified nav_shape greater than file nav size')
        else:
            self._nav_shape = nav_shape
        # nav_shape product is either manual nav_shape if supplied or metadata nav_shape (?)
        self._nav_shape_product = int(prod(self._nav_shape))
        sig_size = int(prod(sig_shape))
        if self._sig_shape is not None:
            manual_sig_size = int(prod(self._sig_shape))
            if (manual_sig_size * self._nav_shape_product) > (self._image_count * sig_size):
                raise DataSetException('Specified sig_shape and nav size '
                                       'too large for data in file')
        else:
            self._sig_shape = sig_shape

        # regardless of file order the Dataset shape property is 'standard'
        shape = Shape(self._nav_shape + self._sig_shape, sig_dims=sig_dims)
        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=shape,
            raw_dtype=np.dtype(self._raw_dtype),
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        return self

    def _get_fileset(self):
        return DMFileSet([
            DMFile(
                path=self._path,
                start_idx=0,
                end_idx=self._image_count,
                sig_shape=self.shape.sig,
                native_dtype=self.meta.raw_dtype,
                file_header=self._array_offset,
            )
        ])

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield DMPartition(
                meta=self.meta,
                partition_slice=part_slice,
                fileset=fileset,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )


class DMFile(File):
    ...


class DMPartition(BasePartition):
    ...



from typing import Any

from libertem.executor.base import AsyncAdapter
from libertem.executor.dask import DaskJobExecutor, cluster_spec
from libertem.utils.devices import detect


def _int_or_zero(value) -> int:
    try:
        return int(value)
    except ValueError:
        return 0


def _convert_device_map(raw_cudas: dict[int, Any]) -> list[int]:
    return [
        this_id
        for dev_id, num in raw_cudas.items()
        for this_id in [dev_id]*_int_or_zero(num)
    ]


def create_executor(*, connection, local_directory, preload) -> DaskJobExecutor:
    devices = detect()
    options = {
        "local_directory": local_directory,
    }
    if "numWorkers" in connection:
        num_workers = connection["numWorkers"]
        if not isinstance(num_workers, int) or num_workers < 1:
            raise ValueError('Number of workers must be positive integer')
        devices["cpus"] = range(num_workers)
    raw_cudas = connection.get("cudas", {})
    cudas = _convert_device_map(raw_cudas)
    devices["cudas"] = cudas
    return DaskJobExecutor.make_local(
        spec=cluster_spec(
            **devices,
            options=options,
            preload=preload,
        )
    )


def create_executor_external(
    executor_spec: dict[str, int],
    local_directory,
    preload,
) -> tuple[AsyncAdapter, dict[str, dict[str, Any]]]:
    cudas = {}
    if executor_spec['cudas']:
        cudas[0] = executor_spec['cudas']
    params = {
        "connection": {
            "type": "LOCAL",
            "numWorkers": executor_spec['cpus'],
            "cudas": cudas,
        }
    }
    sync_executor = create_executor(
        connection=params['connection'],
        local_directory=local_directory,
        preload=preload
    )
    pool = AsyncAdapter.make_pool()
    executor = AsyncAdapter(wrapped=sync_executor, pool=pool)
    return executor, params



import os

import pytest


@pytest.mark.dist
def test_sum_on_dist(raw_on_workers_ipy, ipy_ctx):
    print(ipy_ctx.executor.run_each_host(lambda: os.system("hostname")))
    print(ipy_ctx.executor.get_available_workers().group_by_host())
    print(ipy_ctx.executor.get_available_workers())
    print(ipy_ctx.executor.run_each_host(
        lambda: os.listdir(os.path.dirname(raw_on_workers_ipy._path))))
    analysis = ipy_ctx.create_sum_analysis(dataset=raw_on_workers_ipy)
    results = ipy_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)



import numpy as np

from libertem.contrib.daskadapter import make_dask_array
from libertem.api import Context
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


def test_dask_array():
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(16, 16, 16),
        num_partitions=2,
    )
    (da, workers) = make_dask_array(dataset)
    assert np.allclose(
        da.compute(workers=workers, scheduler='single-threaded'),
        data
    )
    assert np.allclose(
        da.sum().compute(workers=workers, scheduler='single-threaded'),
        data.sum()
    )
    assert da.shape == data.shape


def test_dask_array_2(dask_executor):
    # NOTE: keep in sync with the example in docs/source/api.rst!
    # Construct a Dask array from the dataset
    # The second return value contains information
    # on workers that hold parts of a dataset in local
    # storage to ensure optimal data locality
    ctx = Context(executor=dask_executor)
    dataset = ctx.load("memory", datashape=(16, 16, 16), sig_dims=2)
    dask_array, workers = make_dask_array(dataset)

    # Use the Dask.distributed client of LiberTEM, since it may not be
    # the default client:
    ctx.executor.client.compute(
        dask_array.sum(axis=(-1, -2))
    ).result()


def test_dask_array_with_roi_1():
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(16, 16, 16),
        num_partitions=2,
    )
    roi = np.zeros(dataset.shape.nav, dtype=bool)
    roi[0, 0] = True
    (da, workers) = make_dask_array(dataset, roi=roi)
    assert np.allclose(
        da.compute(workers=workers, scheduler='single-threaded'),
        data[0, 0]
    )
    assert da.shape == (1, 16, 16)


def test_dask_array_with_roi_2():
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(16, 16, 16),
        num_partitions=2,
    )

    sparse_roi = np.random.choice([True, False], size=dataset.shape.nav, p=[0.1, 0.9])
    (da, workers) = make_dask_array(dataset, roi=sparse_roi)
    assert np.allclose(
        da.compute(workers=workers, scheduler='single-threaded'),
        data[sparse_roi]
    )
    assert da.shape == (np.count_nonzero(sparse_roi), 16, 16)



import os
import sys
import logging
import asyncio
import signal
import socket
import select
import threading
import webbrowser
import ipaddress
import urllib.parse
from functools import partial
from typing import Optional
import hmac
import hashlib
import time

import tornado.web
import tornado.gen
import tornado.websocket
import tornado.ioloop
import tornado.escape
from tornado import httputil

from ..common.async_utils import adjust_event_loop_policy
from .shutdown import ShutdownHandler
from .state import SharedState, ExecutorState
from .config import ConfigHandler, ClusterDetailHandler
from .dataset import DataSetDetailHandler, DataSetDetectHandler
from .browse import LocalFSBrowseHandler, LocalFSStatHandler
from .jobs import JobDetailHandler
from .events import ResultEventHandler, EventRegistry
from .connect import ConnectHandler
from .analysis import (
    AnalysisDetailHandler, DownloadDetailHandler, CompoundAnalysisHandler,
    AnalysisRPCHandler,
)
from .generator import DownloadScriptHandler, CopyScriptHandler
from .event_bus import EventBus, MessagePump

log = logging.getLogger(__name__)


class IndexHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    def get(self):
        self.render("client/index.html")


def is_localhost(host):
    return host in ['localhost', '127.0.0.1', '::1']


def _get_token(request):
    token = request.query_arguments.get('token', [b''])[0].decode("utf-8")
    if not token:
        token = request.headers.get('X-Api-Key', '')
    return token


def _write_token_mismatch_error(request):
    start_line = httputil.ResponseStartLine("", 400, "bad request: token mismatch")
    headers = httputil.HTTPHeaders({
        "Content-Type": "text/html; charset=UTF-8",
        "Date": httputil.format_timestamp(time.time()),
    })
    return request.connection.write_headers(
        start_line, headers, b""
    )


def check_token_auth_middleware(app, expected_token):
    # bypass token check if no token is set:
    if expected_token is None:
        return lambda r: app(r)

    expected_hash = hashlib.sha256(expected_token.encode("utf8")).hexdigest()

    def _middleware(request):
        # NOTE: token length may be leaked here
        given_token = _get_token(request)
        given_hash = hashlib.sha256(given_token.encode("utf8")).hexdigest()
        if not hmac.compare_digest(given_hash, expected_hash):
            return _write_token_mismatch_error(request)
        return app(request)

    return _middleware


def make_app(event_registry, shared_state, token=None):
    """
    Returns the fully assembled web API app, which is a
    callable(request) (not the "raw" tornado Application
    instance, because we want to apply middleare around it)
    """
    settings = {
        "static_path": os.path.join(os.path.dirname(__file__), "client"),
    }
    assets_path = os.path.join(os.path.dirname(__file__), "client", "assets")
    common_kwargs = {
        "state": shared_state,
        "event_registry": event_registry,
    }
    app = tornado.web.Application([
        (r"/", IndexHandler, common_kwargs),
        (r"/api/datasets/detect/", DataSetDetectHandler, common_kwargs),
        (r"/api/datasets/([^/]+)/", DataSetDetailHandler, common_kwargs),
        (r"/api/browse/localfs/", LocalFSBrowseHandler, common_kwargs),
        (r"/api/browse/localfs/stat/", LocalFSStatHandler, common_kwargs),
        (r"/api/jobs/([^/]+)/", JobDetailHandler, common_kwargs),
        (r"/api/compoundAnalyses/([^/]+)/analyses/([^/]+)/", AnalysisDetailHandler, common_kwargs),
        (
            r"/api/compoundAnalyses/([^/]+)/analyses/([^/]+)/download/([^/]+)/",
            DownloadDetailHandler,
            common_kwargs
        ),
        (r"/api/compoundAnalyses/([^/]+)/rpc/([^/]+)/", AnalysisRPCHandler, common_kwargs),
        (r"/api/compoundAnalyses/([^/]+)/copy/notebook/", CopyScriptHandler, common_kwargs),
        (r"/api/compoundAnalyses/([^/]+)/download/notebook/", DownloadScriptHandler, common_kwargs),
        (r"/api/compoundAnalyses/([^/]+)/", CompoundAnalysisHandler, common_kwargs),
        (r"/api/events/", ResultEventHandler, common_kwargs),
        (r"/api/shutdown/", ShutdownHandler, common_kwargs),
        (r"/api/config/", ConfigHandler, common_kwargs),
        (r"/api/config/cluster/", ClusterDetailHandler, common_kwargs),
        (r"/api/config/connection/", ConnectHandler, common_kwargs),
        (r"/assets/(.*)", tornado.web.StaticFileHandler, {"path": assets_path}),
    ], **settings)
    app = check_token_auth_middleware(app, token)
    return app


async def do_stop(shared_state):
    log.warning("Exiting...")
    shared_state.executor_state.shutdown()
    loop = asyncio.get_event_loop()
    log.debug("stopping event loop")
    loop.stop()


async def nannynanny():
    '''
    Make sure the event loop wakes up regularly.

    This mitigates a strange bug on Windows
    where Ctrl-C is only handled after an event is processed.

    See Issue #356
    '''
    while True:
        await asyncio.sleep(1)


def sig_exit(signum, frame, shared_state):
    log.info("Handling sig_exit...")
    loop = tornado.ioloop.IOLoop.instance()

    loop.add_callback_from_signal(
        lambda: asyncio.ensure_future(do_stop(shared_state))
    )


def main(bound_sockets, event_registry, shared_state, token):
    app = make_app(event_registry, shared_state, token)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.add_sockets(bound_sockets)
    return http_server


def _confirm_exit(shared_state, loop):
    log.info('interrupted')
    sys.stdout.write("Shutdown libertem server (y/[n])? ")
    sys.stdout.flush()
    r, w, x = select.select([sys.stdin], [], [], 5)
    if r:
        line = sys.stdin.readline()
        if line.lower().startswith('y'):
            log.critical("Shutdown confirmed")
            # schedule stop on main thread
            loop.add_callback_from_signal(
                lambda: asyncio.ensure_future(do_stop(shared_state))
            )
            return
    else:
        print('No answer for 5s: ')
    print('Resuming operation ...')
    # set it back to original SIGINT handler
    loop.add_callback_from_signal(partial(handle_signal, shared_state))


def _handle_exit(signum, frame, shared_state):
    loop = tornado.ioloop.IOLoop.current()
    # register more forceful signal handler for ^C^C case
    signal.signal(signal.SIGINT, partial(sig_exit, shared_state=shared_state))
    thread = threading.Thread(target=partial(_confirm_exit, shared_state, loop))
    thread.daemon = True
    thread.start()


def handle_signal(shared_state):
    if not sys.platform.startswith('win') and sys.stdin and sys.stdin.isatty():
        signal.signal(signal.SIGINT, partial(_handle_exit, shared_state=shared_state))
    else:
        signal.signal(signal.SIGINT, partial(sig_exit, shared_state=shared_state))


def port_from_sockets(*sockets):
    ports = tuple(s.getsockname()[1] for s in sockets)
    assert ports, 'No sockets'
    return ports[0]


def run(
    host, port, browser, local_directory, numeric_level,
    token, preload, strict_port, executor_spec, open_ds,
    snooze_timeout: Optional[float] = None,
):
    logging.basicConfig(
        level=numeric_level,
        format="[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s",
    )

    adjust_event_loop_policy()
    loop = asyncio.get_event_loop()

    # shared state:
    event_bus = EventBus()
    event_registry = EventRegistry()
    pump = MessagePump(event_bus=event_bus, event_registry=event_registry)
    executor_state = ExecutorState(event_bus=event_bus, snooze_timeout=snooze_timeout, loop=loop)
    shared_state = SharedState(executor_state=executor_state)

    executor_state.set_local_directory(local_directory)
    executor_state.set_preload(preload)

    try:
        bound_sockets = tornado.netutil.bind_sockets(port, host)
    except OSError as e:
        if strict_port:
            raise e
        bound_sockets = tornado.netutil.bind_sockets(0, host)
        _port = port_from_sockets(*bound_sockets)
        log.info(f"port {port} already in use, using random open port {_port}")
        port = _port

    try:
        main(bound_sockets, event_registry, shared_state, token)

        async def create_and_set_executor():
            if executor_spec is not None:
                await shared_state.create_and_set_executor(executor_spec)

        try:
            is_ipv6 = isinstance(ipaddress.ip_address(host), ipaddress.IPv6Address)
        except ValueError:
            is_ipv6 = False
        url = f'http://[{host}]:{port}' if is_ipv6 else f'http://{host}:{port}'
        if open_ds is not None:
            url = f'{url}/#action=open&path={open_ds}'
        msg = f"""

    LiberTEM listening on {url}"""
        parts = urllib.parse.urlsplit(url)
        if parts.hostname in ('0.0.0.0', '::'):
            hostname = socket.gethostname()
            mod_url = parts._replace(netloc=f'{hostname}:{parts.port}')
            msg = msg + f"""
                        {urllib.parse.urlunsplit(mod_url)}
    """
        else:
            # For display consistency
            msg = msg + "\n"
        log.info(msg)
        if browser:
            webbrowser.open(url)
        handle_signal(shared_state)
        asyncio.ensure_future(pump.run())
        # Strictly necessary only on Windows, but doesn't do harm in any case.
        # FIXME check later if the unknown root cause was fixed upstream
        asyncio.ensure_future(nannynanny())
        asyncio.ensure_future(create_and_set_executor())
        loop.run_forever()
    finally:
        executor_state.shutdown()



import asyncio

import tornado.websocket

from libertem.web.engine import JobEngine

from .base import log_message
from .messages import Message
from .state import SharedState


class EventRegistry:
    def __init__(self):
        self.handlers = []

    def add_handler(self, handler):
        self.handlers.append(handler)

    def remove_handler(self, handler):
        self.handlers.remove(handler)

    def broadcast_event(self, message, *args, **kwargs):
        futures = []
        for handler in self.handlers:
            try:
                future = handler.write_message(message, *args, **kwargs)
            except tornado.websocket.WebSocketClosedError:
                self.remove_handler(handler)
                continue
            futures.append(future)
        return asyncio.gather(*futures)

    def broadcast_together(self, messages, *args, **kwargs):
        futures = []
        for handler in self.handlers:
            for message in messages:
                futures.append(
                    handler.write_message(message, *args, **kwargs)
                )
        return asyncio.gather(*futures)


class ResultEventHandler(tornado.websocket.WebSocketHandler):
    def initialize(self, state: SharedState, event_registry: EventRegistry):
        self.event_registry = event_registry
        self.state = state
        self.engine = JobEngine(state, event_registry)

    def check_origin(self, origin):
        # FIXME: implement this when we want to support CORS later
        return super().check_origin(origin)

    async def open(self):
        self.event_registry.add_handler(self)
        if self.state.executor_state.have_executor():
            await self.state.dataset_state.verify()
            datasets = await self.state.dataset_state.serialize_all()
            msg = Message().initial_state(
                jobs=self.state.job_state.serialize_all(),
                datasets=datasets, analyses=self.state.analysis_state.serialize_all(),
                compound_analyses=self.state.compound_analysis_state.serialize_all(),
            )
            log_message(msg)
            # FIXME: don't broadcast, only send to the new connection
            self.event_registry.broadcast_event(msg)
            await self.engine.send_existing_job_results()

    def on_close(self):
        self.event_registry.remove_handler(self)



import time
import logging
import asyncio
from typing import TYPE_CHECKING, Callable, Optional, TypeVar

from opentelemetry import trace

from libertem.analysis.base import Analysis, AnalysisResultSet
from libertem.common.executor import JobCancelledError
from libertem.common.tracing import TracedThreadPoolExecutor
from libertem.common.progress import ProgressReporter, ProgressState
from libertem.udf.base import UDFResults
from libertem.io.dataset.base.dataset import DataSet
from libertem.web.models import AnalysisDetails
from .messages import Message
from .state import SharedState
from .base import log_message, result_images

log = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)


if TYPE_CHECKING:
    from .events import EventRegistry

T = TypeVar('T')


class WebProgressReporter(ProgressReporter):
    """
    Progress reporter that pumps `ProgressState`
    messages out via the event registry mechanism.

    The `start`/`update`/`end` callbacks are invoked from
    the context of the `UDFRunner` implementation, so
    we need to make sure we schedule the websocket
    response tasks on the correct event loop
    (the one used by the web API)
    """

    def __init__(
        self,
        state: SharedState,
        event_registry: "EventRegistry",
        loop: asyncio.AbstractEventLoop,
        job_id: str,
    ):
        self.event_registry = event_registry
        self.state = state
        self.loop = loop
        self.job_id = job_id

    def start(self, state: ProgressState):
        self.handle_state_update(state, "start")

    def update(self, state: ProgressState):
        self.handle_state_update(state, "update")

    def end(self, state: ProgressState):
        self.handle_state_update(state, "end")

    def handle_state_update(self, state: ProgressState, event: str):
        msg = Message().job_progress(
            job_id=self.job_id, state=state, event=event
        )

        async def _task():
            await self.event_registry.broadcast_event(msg)
        asyncio.run_coroutine_threadsafe(_task(), loop=self.loop)


class JobEngine:
    def __init__(self, state: SharedState, event_registry: "EventRegistry"):
        self.state = state
        self.event_registry = event_registry
        self._pool = TracedThreadPoolExecutor(tracer=tracer)

    async def run_sync(self, fn: Callable[..., T], *args, **kwargs) -> T:
        # to make sure everything is using `self._pool`, only import
        # `sync_to_async` as a local:
        from libertem.common.async_utils import sync_to_async
        return await sync_to_async(fn, self._pool, *args, **kwargs)

    async def run_analysis(self, analysis_id: str, job_id: str):
        with tracer.start_as_current_span("JobEngine.run_analysis"):
            analysis_state = self.state.analysis_state[analysis_id]
            ds = self.state.dataset_state[analysis_state['dataset']]

            analysis_details = analysis_state["details"]
            analysis_type = analysis_details["analysisType"]
            params = analysis_details["parameters"]

            analysis = Analysis.get_analysis_by_type(analysis_type)(
                dataset=ds,
                parameters=params,
            )

            try:
                if analysis.TYPE != 'UDF':
                    raise TypeError(
                        'Only Analysis classes with TYPE="UDF" are supported'
                    )
                # Users of this class may call `register_job` before actually running it,
                # for example if they want to finish their response before starting to
                # actually run the UDF, so we check here and `register_job` if it hasn't
                # happened already:
                if job_id not in self.state.job_state:
                    await self.register_job(analysis_id, job_id)
                return await self.run_udf(
                    job_id=job_id,
                    dataset=ds,
                    dataset_id=analysis_state['dataset'],
                    analysis=analysis,
                    analysis_id=analysis_id,
                    details=analysis_details,
                )
            except JobCancelledError:
                msg = Message().cancel_done(job_id)
                log_message(msg)
                await self.event_registry.broadcast_event(msg)
                return
            except Exception as e:
                log.exception("error running job, params=%r", params)
                msg = Message().job_error(job_id, "error running job: %s" % str(e))
                self.event_registry.broadcast_event(msg)
                await self.state.job_state.remove(job_id)

    async def register_job(self, analysis_id: str, job_id: str):
        analysis_state = self.state.analysis_state[analysis_id]
        # FIXME: naming? job_state for UDFs?
        self.state.job_state.register(
            job_id=job_id, analysis_id=analysis_id, dataset_id=analysis_state['dataset'],
        )
        self.state.analysis_state.add_job(analysis_id, job_id)

    async def run_udf(
        self,
        job_id: str,
        dataset: DataSet,
        dataset_id: str,
        analysis: Analysis,
        analysis_id: str,
        details: AnalysisDetails,
    ) -> AnalysisResultSet:
        with self.state.executor_state.keep_alive():
            with tracer.start_as_current_span("JobEngine.run_udf"):
                return await self._run_udf(
                    job_id,
                    dataset,
                    dataset_id,
                    analysis,
                    analysis_id,
                    details,
                )

    async def _run_udf(
        self,
        job_id: str,
        dataset: DataSet,
        dataset_id: str,
        analysis: Analysis,
        analysis_id: str,
        details: AnalysisDetails,
    ) -> AnalysisResultSet:
        udf = analysis.get_udf()
        roi = analysis.get_roi()

        executor = await self.state.executor_state.get_executor()
        serialized_job = self.state.job_state.serialize(job_id)
        msg = Message().start_job(
            serialized_job=serialized_job, analysis_id=analysis_id,
        )
        self.event_registry.broadcast_event(msg)

        try:
            return await analysis.controller(  # type: ignore
                cancel_id=job_id, executor=executor,
                job_is_cancelled=lambda: self.state.job_state.is_cancelled(job_id),
                send_results=lambda results, finished: self.send_results(
                    results, job_id, finished=finished,
                    details=details, analysis_id=analysis_id,
                )
            )
        except NotImplementedError:
            pass

        # short circuit if the parameters only change the visualization
        # (as determined by the analysis via `Analysis.need_rerun`):
        if self.state.analysis_state.have_results(analysis_id):
            old_results = self.state.analysis_state.get_results(analysis_id)
            old_details, _, _, old_udf_results = old_results
            if not analysis.need_rerun(
                old_details["parameters"],
                details["parameters"],
            ):
                results = await self.run_sync(
                    analysis.get_udf_results,
                    udf_results=old_udf_results.buffers[0],
                    roi=roi,
                    damage=old_udf_results.damage
                )
                await self.send_results(
                    results, job_id, analysis_id, details, finished=True,
                    udf_results=old_udf_results,
                )
                return results

        t = time.time()
        post_t = time.time()
        window = 0.3
        # FIXME: allow to set correction data for a dataset via upload and local loading
        corrections = dataset.get_correction_data()
        runner_cls = executor.get_udf_runner()
        progress_reporter = WebProgressReporter(
            state=self.state,
            event_registry=self.event_registry,
            loop=asyncio.get_event_loop(),
            job_id=job_id,
        )
        result_iter = runner_cls(
            udfs=[udf],
            progress_reporter=progress_reporter
        ).run_for_dataset_async(
            dataset, executor,
            roi=roi,
            cancel_id=job_id,
            corrections=corrections,
            progress=True,
        )
        async for udf_results in result_iter:
            window = min(max(window, 2*(t - post_t)), 5)
            if time.time() - t < window:
                continue
            results = await self.run_sync(
                analysis.get_udf_results,
                udf_results=udf_results.buffers[0],
                roi=roi,
                damage=udf_results.damage
            )
            post_t = time.time()
            await self.send_results(
                results, job_id, analysis_id, details, udf_results=udf_results
            )
            # The broadcast might have taken quite some time due to
            # backpressure from the network
            t = time.time()

        if self.state.job_state.is_cancelled(job_id):
            raise JobCancelledError()
        results = await self.run_sync(
            analysis.get_udf_results,
            udf_results=udf_results.buffers[0],
            roi=roi,
            damage=udf_results.damage
        )
        await self.send_results(
            results, job_id, analysis_id, details, finished=True, udf_results=udf_results
        )
        return results

    async def send_results(
        self,
        results: AnalysisResultSet,
        job_id: str,
        analysis_id: str,
        details: AnalysisDetails,
        finished=False,
        udf_results: Optional[UDFResults] = None
    ) -> None:
        if self.state.job_state.is_cancelled(job_id):
            raise JobCancelledError()
        images = await result_images(results, self._pool)
        if self.state.job_state.is_cancelled(job_id):
            raise JobCancelledError()
        if finished:
            serialized_job = self.state.job_state.serialize(job_id)
            msg = Message().finish_job(
                serialized_job=serialized_job,
                num_images=len(results),
                image_descriptions=[
                    {"title": result.title, "desc": result.desc,
                    "includeInDownload": result.include_in_download}
                    for result in results
                ],
            )
            self.state.analysis_state.set_results(
                analysis_id, details, results, job_id, udf_results
            )
        else:
            msg = Message().task_result(
                job_id=job_id,
                num_images=len(results),
                image_descriptions=[
                    {"title": result.title, "desc": result.desc,
                    "includeInDownload": result.include_in_download}
                    for result in results
                ],
            )
        log_message(msg)
        # NOTE: make sure the following broadcast_event messages are sent atomically!
        # (that is: keep the code below synchronous, and only send the messages
        # once the images have finished encoding, and then send all at once)
        futures = []
        futures.append(
            self.event_registry.broadcast_event(msg)
        )
        for image in images:
            raw_bytes = image.read()
            futures.append(
                self.event_registry.broadcast_event(raw_bytes, binary=True)
            )
        await asyncio.gather(*futures)

    async def send_existing_job_results(self):
        results = self.state.analysis_state.get_all_results()
        for analysis_id, (details, result_set, job_id, udf_results) in results:
            serialized_job = self.state.job_state.serialize(job_id)
            await self.event_registry.broadcast_event(
                Message().start_job(
                    serialized_job=serialized_job, analysis_id=analysis_id,
                )
            )
            await self.send_results(
                result_set, job_id, analysis_id, details, finished=True, udf_results=udf_results
            )



import io
import os
import numpy as np
import nbformat
import pytest
from temp_utils import _get_hdf5_params
from libertem.analysis.clust import ClusterAnalysis
from libertem.web.notebook_generator.notebook_generator import notebook_generator
from nbconvert.preprocessors import ExecutePreprocessor
from libertem.executor.base import AsyncAdapter


class ResultContainer:
    async def __call__(self, results, finished):
        self.results = results
        self.finished = finished


@pytest.mark.slow
@pytest.mark.asyncio
async def test_clust_default(
    hdf5_ds_2, tmpdir_factory, inline_executor, local_cluster_url,
):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)

    analysis = [{
            "analysisType": 'CLUST',
            "parameters": {
                    'n_peaks': 42,
                    'n_clust': 7,
                    'cy': 3,
                    'cx': 3,
                    'ri': 1,
                    'ro': 5,
                    'delta': 0.05,
                    'min_dist': 1,
                    }
    }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    data_path = os.path.join(datadir, 'cluster_result.npy')
    results = np.load(data_path)

    executor = AsyncAdapter(wrapped=inline_executor)
    analysis = ClusterAnalysis(dataset=hdf5_ds_2, parameters={
        'n_peaks': 42,
        'n_clust': 7,
        'cy': 3,
        'cx': 3,
        'ri': 1,
        'ro': 5,
        'delta': 0.05,
        'min_dist': 1,
    })

    uuid = 'bd3b39fb-0b34-4a45-9955-339da6501bbb'

    res_container = ResultContainer()

    async def send_results(results, finished):
        pass

    await analysis.controller(
        cancel_id=uuid, executor=executor,
        job_is_cancelled=lambda: False,
        send_results=res_container,
    )
    expected = res_container.results

    assert np.allclose(
        results,
        expected['intensity'].raw_data
    )



from libertem.analysis.sum import SumAnalysis
from libertem.masks import _make_circular_mask
from .base import AnalysisResult, AnalysisResultSet
from .helper import GeneratorHelper

import numpy as np


class SumfftTemplate(GeneratorHelper):

    short_name = "sumfft"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return [
            "from libertem.analysis import SumfftAnalysis",
            "from matplotlib import colors",
        ]

    # FIXME write and include documentation
    def get_docs(self):
        docs = ["# SUM FFT Analysis"]
        return '\n'.join(docs)

    def get_analysis(self):
        temp_analysis = [f"sumfft_analysis = SumfftAnalysis(dataset=ds, parameters={self.params})",
                         "sumfft_result = ctx.run(sumfft_analysis, progress=True)"]
        return '\n'.join(temp_analysis)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(sumfft_result.intensity, norm=colors.LogNorm())",
                "plt.colorbar()"]
        return ['\n'.join(plot)]


class SumfftAnalysis(SumAnalysis, id_="FFTSUM_FRAMES"):
    TYPE = 'UDF'

    def get_udf_results(self, udf_results, roi, damage):
        from libertem.viz import visualize_simple
        sum_results = np.array(udf_results['intensity'])
        real_rad = self.parameters.get("real_rad")
        real_center = (self.parameters.get("real_centery"), self.parameters.get("real_centerx"))

        if not (real_center is None or real_rad is None):
            sigshape = sum_results.shape
            real_mask = 1-1*_make_circular_mask(
                    real_center[1], real_center[0], sigshape[1], sigshape[0], real_rad
                )
            fft_result = np.log(abs(np.fft.fftshift(np.fft.fft2(sum_results*real_mask)))+1)
        else:
            fft_result = np.log(abs(np.fft.fftshift(np.fft.fft2(sum_results)))+1)
        return AnalysisResultSet([
            AnalysisResult(
                raw_data=sum_results,
                visualized=visualize_simple(fft_result, damage=True),
                key="intensity", title="intensity", desc="fft of sum of all frames"
            ),
        ])

    @classmethod
    def get_template_helper(cls):
        return SumfftTemplate



import pytest
import numpy as np
import numba
from sparseconverter import NUMPY, SPARSE_COO, SPARSE_BACKENDS, get_device_class, for_backend

from libertem.udf.stddev import StdDevUDF, run_stddev, process_tile, merge
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random, set_device_class


@pytest.mark.with_numba
@pytest.mark.parametrize(
    "use_roi", [True, False]
)
@pytest.mark.parametrize(
    "dtype", [np.float32, np.complex128]
)
@pytest.mark.parametrize(
    "use_numba", [True, False]
)
@pytest.mark.parametrize(
    'backend', (None, ) + tuple(StdDevUDF().get_backends())
)
def test_stddev(lt_ctx, delayed_ctx, use_roi, dtype, use_numba, backend):
    """
    Test variance, standard deviation, sum of frames, and mean computation
    implemented in udf/stddev.py

    Parameters
    ----------
    lt_ctx
        Context class for loading dataset and creating jobs on them
    """
    with set_device_class(get_device_class(backend)):
        if backend in SPARSE_BACKENDS:
            data = _mk_random(size=(30, 3, 516), dtype=dtype, array_backend=SPARSE_COO)
        else:
            data = _mk_random(size=(30, 3, 516), dtype=dtype, array_backend=NUMPY)
        dataset = MemoryDataSet(
            data=data,
            tileshape=(3, 2, 257),
            num_partitions=2,
            sig_dims=2,
            array_backends=(backend, ) if backend is not None else None
        )
        if use_roi:
            roi = np.random.choice([True, False], size=dataset.shape.nav)
            res = run_stddev(lt_ctx, dataset, roi=roi, use_numba=use_numba)
            res_delayed = run_stddev(delayed_ctx, dataset, roi=roi, use_numba=use_numba)
        else:
            roi = np.ones(dataset.shape.nav, dtype=bool)
            res = run_stddev(lt_ctx, dataset, use_numba=use_numba)
            res_delayed = run_stddev(delayed_ctx, dataset, use_numba=use_numba)

        assert 'sum' in res
        assert 'num_frames' in res
        assert 'var' in res
        assert 'mean' in res
        assert 'std' in res

        N = np.count_nonzero(roi)
        assert res['num_frames'] == N  # check the total number of frames
        assert res_delayed['num_frames'] == N

        print('sum')
        refsum = for_backend(np.sum(data[roi], axis=0), NUMPY)
        print(res['sum'])
        print(refsum)
        print(res['sum'] - refsum)
        assert np.allclose(res['sum'], refsum)  # check sum of frames
        assert np.allclose(res_delayed['sum'], refsum)

        ref_mean = for_backend(np.mean(data[roi], axis=0), NUMPY)
        assert np.allclose(res['mean'], ref_mean)  # check mean
        assert np.allclose(res_delayed['mean'], ref_mean)

        refvar = for_backend(np.var(data[roi], axis=0), NUMPY)
        print('var')
        print(res['var'])
        print(refvar)
        print(refvar - res['var'])
        assert np.allclose(refvar, res['var'])  # check variance
        assert np.allclose(refvar, res_delayed['var'])

        refstd = for_backend(np.std(data[roi], axis=0), NUMPY)
        assert np.allclose(refstd, res['std'])  # check standard deviation
        assert np.allclose(refstd, res_delayed['std'])


@pytest.mark.slow
def test_stddev_fuzz(concurrent_ctx):
    for i in range(100):
        total = np.random.randint(1, 20)
        tile = np.random.randint(1, total + 1)
        n_part = np.random.randint(1, total//tile + 1)
        print(total, tile, n_part)
        data = np.random.randn(total, 512, 512) + 1j*np.random.randn(total, 512, 512)
        ds = concurrent_ctx.load(
            'memory',
            data=data,
            tileshape=(tile, 512, 512),
            num_partitions=n_part,
            sig_dims=2,
        )
        var = np.var(data, axis=0)
        res = concurrent_ctx.run_udf(dataset=ds, udf=StdDevUDF())
        assert np.allclose(res['var'], var)


@numba.njit(boundscheck=True, cache=True)
def _stability_workhorse(data):
    # This function imitates the calculation flow of a single pixel for a very big dataset.
    # The partition and tile structure is given by the shape of data
    # data.shape[0]: partitions
    # data.shape[1]: tiles per partition
    # data.shape[2]: frames per tile
    # data.shape[3]: mock signal shape, can be 1
    # The test uses numba since this allows efficient calculations for small units
    # of data. Running it with JIT disabled is very, very slow!
    s = np.zeros(1, dtype=np.float32)
    varsum = np.zeros(1, dtype=np.float32)
    N = 0
    partitions = data.shape[0]
    tiles = data.shape[1]
    frames = data.shape[2]
    for partition in range(partitions):
        partition_sum = np.zeros(1, dtype=np.float32)
        partition_varsum = np.zeros(1, dtype=np.float32)
        partition_N = 0
        for tile in range(tiles):
            if partition_N == 0:
                partition_sum[:] = np.sum(data[partition, tile])
                partition_varsum[:] = np.var(data[partition, tile]) * frames
                partition_N = frames
            else:
                partition_N = process_tile(
                    tile=data[partition, tile],
                    n_0=partition_N,
                    sum_inout=partition_sum,
                    varsum_inout=partition_varsum
                )
        N = merge(
            dest_n=N,
            dest_sum=s,
            dest_varsum=varsum,
            src_n=partition_N,
            src_sum=partition_sum,
            src_varsum=partition_varsum,
            src_mean=partition_sum / partition_N,
        )
    return N, s, varsum


# This shouldn't be @pytest.mark.numba
# since the calculation will take forever
# with JIT disabled
@pytest.mark.slow
def test_stability(lt_ctx):
    data = _mk_random(size=(1024, 1024, 8, 1), dtype="float32")

    N, s, varsum = _stability_workhorse(data)

    assert N == np.prod(data.shape)
    assert np.allclose(data.sum(), s)
    assert np.allclose(data.var(ddof=N-1), varsum)



import inspect

import numpy as np
import sparse

from libertem.udf.raw import PickUDF
from .base import BaseAnalysis, AnalysisResult, AnalysisResultSet
from .helper import GeneratorHelper


class PickTemplate(GeneratorHelper):

    short_name = "pick"
    api = "create_pick_analysis"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return ["from matplotlib import colors"]

    def get_docs(self):
        title = "Pick Analysis"
        from libertem.api import Context
        docs_rst = inspect.getdoc(Context.create_pick_analysis)
        docs = self.format_docs(title, docs_rst)
        return docs

    def convert_params(self):
        params = ['dataset=ds']
        for k in ['x', 'y']:
            params.append(f'{k}={self.params[k]}')
        return ', '.join(params)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(pick_result['intensity'], norm=colors.LogNorm())",
                "plt.colorbar()"]
        return ['\n'.join(plot)]


class PickResultSet(AnalysisResultSet):
    """
    Running a :class:`PickFrameAnalysis` via :meth:`libertem.api.Context.run`
    returns an instance of this class.

    If the dataset contains complex numbers, the regular result attribute carries the
    absolute value of the result and additional attributes with real part, imaginary part,
    phase and full complex result are available.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.4.0
        Picking now returns data in the native dtype of the dataset with the new UDF back-end.

    Attributes
    ----------
    intensity : libertem.analysis.base.AnalysisResult
        The specified detector frame. Absolute value if the dataset
        contains complex numbers. Log-scaled visualization.
    intensity_lin : libertem.analysis.base.AnalysisResult
        The specified detector frame. Absolute value if the dataset
        contains complex numbers. Linear visualization.

        .. versionadded:: 0.6.0

    intensity_real : libertem.analysis.base.AnalysisResult
        Real part of the specified detector frame. This is only available if the dataset
        contains complex numbers.
    intensity_imag : libertem.analysis.base.AnalysisResult
        Imaginary part of the specified detector frame. This is only available if the dataset
        contains complex numbers.
    intensity_angle : libertem.analysis.base.AnalysisResult
        Phase angle of the specified detector frame. This is only available if the dataset
        contains complex numbers.
    intensity_complex : libertem.analysis.base.AnalysisResult
        Complex result of the specified detector frame. This is only available if the dataset
        contains complex numbers.
    """
    pass


class PickFrameAnalysis(BaseAnalysis, id_="PICK_FRAME"):
    TYPE = 'UDF'
    """
    Pick a single, complete frame from a dataset
    """

    def get_origin(self):
        dims = self.dataset.shape.nav.dims
        if dims not in (1, 2, 3):
            raise ValueError(
                "can only handle 1D/2D/3D nav currently, received %s dimensions" % dims
            )
        zyx = (
            self.parameters.get('z'),
            self.parameters.get('y'),
            self.parameters.get('x'),
        )
        messages = {
            1: "Need x, not y and not z to index 1D dataset, received z=%s, y=%s, x=%s",
            2: "Need x, y and not z to index 2D dataset, received z=%s, y=%s, x=%s",
            3: "Need x, y z to index 3D dataset, received z=%s, y=%s, x=%s",
        }
        keep = zyx[-dims:]
        drop = zyx[:-dims]

        if (None in keep) or not all(d is None for d in drop):
            raise ValueError(messages[dims] % zyx)

        return keep

    def get_udf(self):
        return PickUDF()

    def get_roi(self):
        coords = np.array(self.get_origin())[:, np.newaxis]
        roi = sparse.COO(coords=coords, data=True, fill_value=False, shape=self.dataset.shape.nav)
        return roi

    def get_udf_results(self, udf_results, roi, damage):
        return self.get_generic_results(udf_results['intensity'].data[0], damage=True)

    def get_coords(self):
        parameters = self.parameters
        coords = [
            "%s=%d" % (axis, parameters.get(axis))
            for axis in ['x', 'y', 'z']
            if axis in parameters
        ]
        return " ".join(coords)

    def get_generic_results(self, data, damage):
        from libertem.viz import visualize_simple
        coords = self.get_coords()

        if data.dtype.kind == 'c':
            return AnalysisResultSet(
                self.get_complex_results(
                    data,
                    key_prefix="intensity",
                    title="intensity",
                    desc=f"the frame at {coords}",
                    damage=True,
                    default_lin=False,
                )
            )
        return AnalysisResultSet([
            AnalysisResult(
                raw_data=data,
                visualized=visualize_simple(data, logarithmic=True, damage=True),
                key="intensity", title="intensity [log]",
                desc=f"the frame at {coords} log-scaled"
            ),
            AnalysisResult(
                raw_data=data,
                visualized=visualize_simple(data, logarithmic=False, damage=True),
                key="intensity_lin", title="intensity [lin]",
                desc=f"the frame at {coords} lin-scaled"
            ),
        ])

    @classmethod
    def get_template_helper(cls):
        return PickTemplate



import time
import logging
from typing import Optional, TYPE_CHECKING
from collections.abc import Sequence

import psutil
import numpy as np
from sparseconverter import (
    BACKENDS, ArrayBackend, for_backend, get_backend, SPARSE_BACKENDS, get_device_class,
    conversion_cost
)

from libertem.common.math import prod, count_nonzero, flat_nonzero
from libertem.common.messageconverter import MessageConverter
from libertem.io.dataset.base import (
    FileSet, BasePartition, DataSet, DataSetMeta, TilingScheme,
    File, MMapBackend, DataSetException
)
from libertem.io.dataset.base.backend_mmap import MMapBackendImpl, MMapFile
from libertem.common import Shape, Slice
from libertem.io.dataset.base import DataTile

if TYPE_CHECKING:
    from libertem.common.executor import JobExecutor

log = logging.getLogger(__name__)


class FakeMMapFile(MMapFile):
    """
    Implementing the same interface as MMapFile, without filesystem backing
    """
    def open(self):
        self._arr = self.desc._data
        self._mmap = self.desc._data
        return self

    def close(self):
        self._arr = None
        self._mmap = None


class MemBackend(MMapBackend):
    def get_impl(self):
        return MemBackendImpl()


class MemBackendImpl(MMapBackendImpl):
    FILE_CLS = FakeMMapFile

    def _set_readahead_hints(self, roi, fileset):
        pass

    def _get_tiles_roi(self, tiling_scheme, open_files, read_ranges, roi, sync_offset):
        ds_sig_shape = tiling_scheme.dataset_shape.sig
        sig_dims = tiling_scheme.shape.sig.dims
        slices, ranges, scheme_indices = read_ranges

        fh = open_files[0]
        memmap = fh.array.reshape((fh.desc.num_frames,) + tuple(ds_sig_shape))
        flat_roi = np.full(roi.reshape((-1,)).shape, False)
        roi_nonzero = flat_nonzero(roi)
        offset_roi = np.clip(roi_nonzero + sync_offset, 0, flat_roi.size)
        flat_roi[offset_roi] = True
        data_w_roi = memmap[flat_roi]

        for idx in range(slices.shape[0]):
            origin, shape = slices[idx]
            scheme_idx = scheme_indices[idx]

            tile_slice = Slice(
                origin=origin,
                shape=Shape(shape, sig_dims=sig_dims)
            )
            if sync_offset >= 0:
                data_slice = tile_slice.get()
            else:
                frames_to_skip = count_nonzero(roi.reshape((-1,))[:abs(sync_offset)])
                data_slice = Slice(
                    origin=(origin[0] - frames_to_skip,) + tuple(origin[-sig_dims:]),
                    shape=Shape(shape, sig_dims=sig_dims)
                )
                data_slice = data_slice.get()
            data = data_w_roi[data_slice]
            yield DataTile(
                data,
                tile_slice=tile_slice,
                scheme_idx=scheme_idx,
            )

    def get_tiles(
        self, decoder, tiling_scheme, fileset, read_ranges, roi, native_dtype, read_dtype,
        sync_offset, corrections, array_backend: ArrayBackend,
    ):
        if roi is None:
            # support arbitrary tiling in case of no roi
            with self.open_files(fileset) as open_files:
                if sync_offset >= 0:
                    for tile in self._get_tiles_straight(
                        tiling_scheme, open_files, read_ranges, sync_offset
                    ):
                        if tile.dtype != read_dtype or tile.c_contiguous is False:
                            data = tile.data.astype(read_dtype)
                        else:
                            data = tile.data
                        self.preprocess(data, tile.tile_slice, corrections)
                        data = for_backend(data, array_backend)
                        yield DataTile(data, tile.tile_slice, tile.scheme_idx)
                else:
                    for tile in self._get_tiles_w_copy(
                        tiling_scheme=tiling_scheme,
                        open_files=open_files,
                        read_ranges=read_ranges,
                        read_dtype=read_dtype,
                        native_dtype=native_dtype,
                        decoder=decoder,
                        corrections=corrections,
                    ):
                        data = for_backend(tile.data, array_backend)
                        yield DataTile(data, tile.tile_slice, tile.scheme_idx)
        else:
            with self.open_files(fileset) as open_files:
                for tile in self._get_tiles_roi(
                    tiling_scheme=tiling_scheme,
                    open_files=open_files,
                    read_ranges=read_ranges,
                    roi=roi,
                    sync_offset=sync_offset,
                ):
                    data = tile.data.astype(read_dtype)
                    self.preprocess(data, tile.tile_slice, corrections)
                    data = for_backend(data, array_backend)
                    yield DataTile(data, tile.tile_slice, tile.scheme_idx)


class MemDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/MEMDatasetParams.schema.json",
        "title": "MEMDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "MEMORY"},
            "tileshape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
            },
            "datashape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
            },
            "num_partitions": {"type": "number", "minimum": 1},
            "sig_dims": {"type": "number", "minimum": 1},
            "check_cast": {"type": "boolean"},
            "crop_frames": {"type": "boolean"},
            "tiledelay": {"type": "number"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "array_backend": {"type": "string"},
        },
        "required": ["type", "tileshape", "num_partitions"],
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["tileshape", "num_partitions", "sig_dims", "check_cast",
                      "crop_frames", "tiledelay", "datashape", "array_backend"]
            if k in raw_data
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class MemoryFile(File):
    def __init__(self, data, check_cast=True, *args, **kwargs):
        self._data = data
        self._check_cast = check_cast
        super().__init__(*args, **kwargs)

    @property
    def data(self):
        return self._data


class MemoryDataSet(DataSet):
    '''
    This dataset is constructed from a NumPy array in memory for testing
    purposes. It is not recommended for production use since it performs poorly with a
    distributed executor.

    Examples
    --------

    >>> data = np.zeros((2, 2, 64, 64), dtype=np.float32)
    >>> ds = ctx.load('memory', data=data, sig_dims=2)
    '''
    def __init__(self, tileshape=None, num_partitions=None, data=None, sig_dims=None,
                 check_cast=True, tiledelay=None, datashape=None, base_shape=None,
                 force_need_decode=False, io_backend=None,
                 nav_shape=None, sig_shape=None, sync_offset=0, array_backends=None):
        super().__init__(io_backend=io_backend)
        if io_backend is not None:
            raise ValueError("MemoryDataSet currently doesn't support alternative I/O backends")
        # For HTTP API testing purposes: Allow to create empty dataset with given shape
        if data is None:
            if datashape is None:
                raise DataSetException('MemoryDataSet can be created from either data [np.ndarray],'
                                       ' or datashape [tuple | Shape], both arguments are None')
            data = np.zeros(datashape, dtype=np.float32)
        if num_partitions is None:
            num_partitions = psutil.cpu_count(logical=False)
        # if tileshape is None:
        #     sig_shape = data.shape[-sig_dims:]
        #     target = 2**20
        #     framesize = prod(sig_shape)
        #     framecount = max(1, min(prod(data.shape[:-sig_dims]), int(target / framesize)))
        #     tileshape = (framecount, ) + sig_shape
        self.data = data
        self._base_shape = base_shape
        self.num_partitions = num_partitions

        if sig_dims is None:
            if sig_shape is not None:
                sig_dims = len(sig_shape)
            elif nav_shape is not None:
                sig_dims = len(data.shape) - len(nav_shape)
            else:
                sig_dims = 2
        else:
            if sig_shape is not None and len(sig_shape) != sig_dims:
                raise ValueError(
                    f"Length of sig_shape {sig_shape} not matching sig_dims {sig_dims}."
                )

        self.sig_dims = sig_dims
        if nav_shape is None:
            nav_shape = data.shape[:-sig_dims]
        else:
            nav_shape = tuple(nav_shape)
        if sig_shape is None:
            sig_shape = data.shape[-sig_dims:]
        else:
            sig_shape = tuple(sig_shape)
        if self.data.size % prod(sig_shape) != 0:
            raise ValueError("Data size is not a multiple of sig shape")
        self._image_count = self.data.size // prod(sig_shape)
        self._nav_shape = nav_shape
        self._sig_shape = sig_shape
        self._sync_offset = sync_offset
        self._array_backends = array_backends
        self._check_cast = check_cast
        self._tiledelay = tiledelay
        self._force_need_decode = force_need_decode
        self._nav_shape_product = int(prod(nav_shape))
        self._shape = Shape(
            nav_shape + sig_shape, sig_dims=self.sig_dims
        )
        if tileshape is None:
            self.tileshape = None
        else:
            assert len(tileshape) == self.sig_dims + 1
            self.tileshape = Shape(tileshape, sig_dims=self.sig_dims)
        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=self._shape,
            array_backends=self.array_backends,
            raw_dtype=self.data.dtype,
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )

    def initialize(self, executor):
        return self

    @classmethod
    def get_msg_converter(cls):
        return MemDatasetParams

    @classmethod
    def detect_params(cls, data: np.ndarray, executor: "JobExecutor"):
        try:
            _ = data.shape
            return {
                "parameters": {
                    "data": data,
                },
                "info": {}
            }
        except AttributeError:
            return False

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def shape(self):
        return self._shape

    @property
    def array_backends(self) -> Sequence[ArrayBackend]:
        """
        All backends can be returned on request

        .. versionadded:: 0.11.0
        """
        if self._array_backends is None:
            native = get_backend(self.data)
            is_sparse = native in SPARSE_BACKENDS
            native_device_class = get_device_class(native)
            cost_metric = {}
            # Sort by tuple (cost_override, conversion_cost),
            # meaning preference for native backend, same sparsity
            # and same device class take precedence over measured conversion cost
            for backend in BACKENDS:
                cost_metric[backend] = [5, conversion_cost(native, backend)]
                if backend == native:
                    cost_metric[backend][0] -= 2
                # sparse==sparse or dense==dense
                if (backend in SPARSE_BACKENDS) == is_sparse:
                    cost_metric[backend][0] -= 2
                # Same device class
                if get_device_class(backend) == native_device_class:
                    cost_metric[backend][0] -= 1
            return tuple(sorted(BACKENDS, key=lambda k: cost_metric[k]))
        else:
            return self._array_backends

    def check_valid(self):
        return True

    def get_cache_key(self):
        return TypeError("memory data set is not cacheable yet")

    def get_num_partitions(self):
        return self.num_partitions

    def get_base_shape(self, roi):
        if self.tileshape is not None:
            return self.tileshape
        if self._base_shape is not None:
            return self._base_shape
        return super().get_base_shape(roi)

    def adjust_tileshape(
        self, tileshape: tuple[int, ...], roi: Optional[np.ndarray],
    ) -> tuple[int, ...]:
        if self.tileshape is not None:
            return tuple(self.tileshape)
        return super().adjust_tileshape(tileshape, roi)

    def need_decode(self, read_dtype, roi, corrections):
        if self._force_need_decode:
            return True
        return super().need_decode(read_dtype, roi, corrections)

    def get_io_backend(self):
        return MemBackend()

    def get_partitions(self):
        fileset = FileSet([
            MemoryFile(
                path=None,
                start_idx=0,
                end_idx=self._image_count,
                native_dtype=self.data.dtype,
                sig_shape=self.shape.sig,
                data=self.data.reshape((-1, *self.shape.sig)),
                check_cast=self._check_cast,
            )
        ])

        for part_slice, start, stop in self.get_slices():
            log.debug(
                "creating partition slice %s start %s stop %s",
                part_slice, start, stop,
            )
            yield MemPartition(
                meta=self._meta,
                partition_slice=part_slice,
                fileset=fileset,
                start_frame=start,
                num_frames=stop - start,
                tiledelay=self._tiledelay,
                tileshape=self.tileshape,
                force_need_decode=self._force_need_decode,
                io_backend=self.get_io_backend(),
                decoder=self.get_decoder(),
            )


class MemPartition(BasePartition):
    def __init__(self, tiledelay, tileshape, force_need_decode=False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tiledelay = tiledelay
        self._tileshape = tileshape
        self._force_tileshape = True
        self._force_need_decode = force_need_decode

    def get_io_backend(self):
        return MemBackend()

    def get_macrotile(self, *args, **kwargs):
        self._force_tileshape = False
        mt = super().get_macrotile(*args, **kwargs)
        self._force_tileshape = True
        return mt

    def get_tiles(self, *args, **kwargs):
        if args and isinstance(args[0], TilingScheme):
            tiling_scheme = args[0]
            args = args[1:]
            intent = tiling_scheme.intent
        elif 'tiling_scheme' in kwargs:
            tiling_scheme = kwargs.pop('tiling_scheme')
            intent = tiling_scheme.intent
        else:
            # In this case we require the next if-block to execute
            intent = None
        # force our own tiling_scheme, if a tileshape is given:
        if self._tileshape is not None and self._force_tileshape:
            tiling_scheme = TilingScheme.make_for_shape(
                tileshape=self._tileshape,
                dataset_shape=self.meta.shape,
                intent=intent,
            )
        tiles = super().get_tiles(tiling_scheme, *args, **kwargs)
        if self._tiledelay:
            log.debug("delayed get_tiles, tiledelay=%.3f" % self._tiledelay)
            for tile in tiles:
                yield tile
                time.sleep(self._tiledelay)
        else:
            yield from tiles



import io
import os
import glob

import nbformat
import pytest
import numpy as np
from nbconvert.preprocessors import ExecutePreprocessor

from libertem.web.notebook_generator.notebook_generator import notebook_generator

from utils import get_testdata_path
from temp_utils import _get_hdf5_params, create_random_hdf5


pytestmark = [pytest.mark.slow]

TMP_TESTDATA_PATH = os.path.join(get_testdata_path(), 'temp_data')
HAVE_TMP_TESTDATA = os.path.exists(TMP_TESTDATA_PATH)


def test_ring_default(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)

    analysis = [{
            "analysisType": "APPLY_RING_MASK",
            "parameters": {
                    'shape': 'ring',
                    'cx': 8,
                    'cy': 8,
                    'ri': 5,
                    'ro': 8,
                    }
            }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    data_path = os.path.join(datadir, 'ring_result.npy')
    results = np.load(data_path)

    analysis = lt_ctx.create_ring_analysis(
                            dataset=hdf5_ds_2,
                            cx=8,
                            cy=8,
                            ri=5,
                            ro=8
                        )
    expected = lt_ctx.run(analysis)

    assert np.allclose(
        results,
        expected['intensity'].raw_data,
    )


@pytest.fixture(scope='function')
def random_hdf5_1():
    tmp_dir = os.path.join(TMP_TESTDATA_PATH, 'tmp_ring_template')
    os.mkdir(tmp_dir)
    ds_path = os.path.join(tmp_dir, 'tmp_random_hdf5.h5')
    ds = create_random_hdf5(ds_path)
    yield ds
    os.remove(ds_path)
    os.rmdir(tmp_dir)


@pytest.mark.dist
@pytest.mark.asyncio
@pytest.mark.skipif(not HAVE_TMP_TESTDATA, reason="need temporary directory for testdata")  # NOQA
def test_ring_tcp_cluster(lt_ctx, random_hdf5_1, scheduler_addr):

    conn = {"connection": {
                    "type": "TCP",
                    "address": scheduler_addr
                    }
            }
    ds = random_hdf5_1
    ds_path = ds.path
    tmp_dir = os.path.dirname(ds_path)
    dataset = _get_hdf5_params(ds_path)

    analysis = [{
            "analysisType": "APPLY_RING_MASK",
            "parameters": {
                    'shape': 'ring',
                    'cx': 8,
                    'cy': 8,
                    'ri': 5,
                    'ro': 8,
                    }
            }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    data_path = None
    try:
        ep.preprocess(nb, {"metadata": {"path": tmp_dir}})
        data_path = os.path.join(tmp_dir, 'ring_result.npy')
        results = np.load(data_path)
    finally:
        if data_path is not None and os.path.exists(data_path):
            os.remove(data_path)

    analysis = lt_ctx.create_ring_analysis(
                            dataset=ds,
                            cx=8,
                            cy=8,
                            ri=5,
                            ro=8
                        )
    expected = lt_ctx.run(analysis)

    assert np.allclose(
        results,
        expected['intensity'].raw_data,
    )


@pytest.mark.dist
def test_datadir_inspection():
    print("get_testdata_path():", get_testdata_path())
    print(list(sorted(glob.glob(get_testdata_path() + '/*'))))
    print(list(sorted(glob.glob(TMP_TESTDATA_PATH + '/*'))))
    assert HAVE_TMP_TESTDATA



import functools
import logging
from typing import Union, Callable, Optional
from collections.abc import Sequence
from typing_extensions import Literal

import sparse
import scipy.sparse
import numpy as np
import numpy.typing as npt
import cloudpickle
from sparseconverter import (
    CPU_BACKENDS, CUDA, CUPY_BACKENDS, for_backend, ArrayT, ArrayBackend, NUMPY
)
from libertem.io.dataset.base.tiling_scheme import TilingScheme

from libertem.common.sparse import to_dense, to_sparse, is_sparse
from libertem.common import Slice


log = logging.getLogger(__name__)

FactoryT = Callable[[], ArrayT]
SparseSupportedT = Literal[
    'sparse.pydata',
    'sparse.pydata.GCXS',
    'scipy.sparse',
    'scipy.sparse.csc',
    'scipy.sparse.csr',
]


def _build_sparse(m, dtype: npt.DTypeLike, sparse_backend: SparseSupportedT, backend: ArrayBackend):
    if sparse_backend == 'sparse.pydata' and backend == NUMPY:
        # sparse.pydata.org is fastest for masks with few layers
        # and few entries
        return m.astype(dtype)
    elif sparse_backend == 'sparse.pydata.GCXS' and backend == NUMPY:
        # sparse.pydata.org is fastest for masks with few layers
        # and few entries
        return sparse.GCXS(m.astype(dtype))
    elif 'scipy.sparse' in sparse_backend:
        if backend in CPU_BACKENDS or backend == CUDA:
            lib = scipy.sparse
        elif backend in CUPY_BACKENDS:
            # Avoid import if possible
            import cupyx.scipy.sparse
            lib = cupyx.scipy.sparse
        else:
            raise ValueError(
                f"Backend {backend} not supported for sparse_backend {sparse_backend}."
            )
        iis, jjs = m.coords
        values = m.data
        if sparse_backend == 'scipy.sparse.csc':
            s = scipy.sparse.csc_matrix(
                (values, (iis, jjs)), shape=m.shape, dtype=dtype)
            assert s.has_canonical_format
            return lib.csc_matrix(s)
        elif sparse_backend == 'scipy.sparse' or sparse_backend == 'scipy.sparse.csr':
            s = scipy.sparse.csr_matrix(
                (values, (iis, jjs)), shape=m.shape, dtype=dtype)
            assert s.has_canonical_format
            return lib.csr_matrix(s)
    # Fall through if no return statement was reached
    raise ValueError(
        f"sparse_backend {sparse_backend} not implemented for backend {backend}. "
        "CPU-based backends supports 'sparse.pydata', 'sparse.pydata.GCXS', 'scipy.sparse', "
        "'scipy.sparse.csc' or 'scipy.sparse.csr'. "
        "CUDA-based backends supports 'scipy.sparse', 'scipy.sparse.csc' or 'scipy.sparse.csr'. "
    )


def _make_mask_slicer(
    computed_masks: ArrayT,
    dtype: npt.DTypeLike,
    sparse_backend: Union[Literal[False], SparseSupportedT],
    transpose: bool,
    backend: ArrayBackend,
):
    @functools.cache
    def _get_masks_for_slice(slice_):
        stack_height = computed_masks.shape[0]
        m = slice_.get(computed_masks, sig_only=True)
        # We need the mask's signal dimension flattened
        m = m.reshape((stack_height, -1))
        if transpose:
            # We need the stack transposed in the next step
            m = m.T
        if sparse_backend is False:
            return for_backend(m, backend).astype(dtype)
        else:
            return _build_sparse(m, dtype, sparse_backend, backend)
    return _get_masks_for_slice


class MaskContainer:
    '''
    Container for mask stacks that are created from factory functions.

    It allows stacking, cached slicing, transposing and conversion
    to condition the masks for high-performance dot products.

    Computation of masks is delayed until as late as possible,
    but is done automatically when necessary. Methods which can trigger
    mask instantiation include:

      - container.use_sparse
      - len(container) [if the count argument is None at __init__]
      - container.dtype [if the dtype argument is None at __init__]
      - any of the get() methods

    use_sparse at init can be None, False, True or any supported
    sparse backend as a string in {'scipy.sparse', 'scipy.sparse.csc',
    'scipy.sparse.csr', 'sparse.pydata', 'sparse.pydata.GCXS'}

    use_sparse as None means the sparse mode will be chosen only after
    the masks are instantiated. All masks being sparse will activate sparse
    processing using the backend in default_sparse, else dense processing
    will be used on the appropriate backend.
    '''
    def __init__(
        self,
        mask_factories: Union[FactoryT, Sequence[FactoryT]],
        dtype: Optional[npt.DTypeLike] = None,
        use_sparse: Optional[Union[bool, SparseSupportedT]] = None,
        count: Optional[int] = None,
        backend: Optional[ArrayBackend] = None,
        default_sparse: SparseSupportedT = 'scipy.sparse',
    ):
        self.mask_factories = mask_factories
        # If we generate a whole mask stack with one function call,
        # we should know the length without generating the mask stack
        self._length = count
        self._dtype = dtype
        self._mask_cache = {}
        # lazily initialized in the worker process, to keep task size small:
        self._computed_masks = None
        if backend is None:
            backend = 'numpy'
        self.backend = backend
        self._get_masks_for_slice = {}
        # from Python 3.8....
        # assert default_sparse in typing.get_args(SparseSupportedT)
        self._default_sparse = default_sparse
        self._use_sparse: Union[Literal[False], None, SparseSupportedT]
        # Try to resolve if we are actually using sparse upfront,
        # this is not always possible as it depends on whether the
        # mask_factories will all return sparse matrices
        if use_sparse is True:
            self._use_sparse = default_sparse
        elif use_sparse is False:
            self._use_sparse = False
        elif isinstance(use_sparse, str) and (
            # This should be rendered compatible with SPARSE_BACKENDS frozenset
            # but there are issues of capitalization and naming
            use_sparse.lower().startswith('scipy.sparse')
            or use_sparse.lower().startswith('sparse.pydata')
        ):
            self._use_sparse = use_sparse
        elif use_sparse is None:
            # User doesn't specify, will use sparse if masks
            # are sparse and we are on a compatible backend
            if (
                default_sparse.startswith('sparse.pydata')
                and self.backend in CUPY_BACKENDS
            ):
                # sparse.pydata cannot run on CuPy, so densify to allow calculation
                self._use_sparse = False
            else:
                # we can't determine _use_sparse without creating the masks
                # themselves and we want to delay this as late as possible
                # leave as None for now and resolve on first access to
                # the self.use_sparse property
                self._use_sparse = None
        else:
            raise ValueError(f'use_sparse not an allowed value: {use_sparse}')

        self.validate_mask_functions()

    def __getstate__(self):
        # don't even try to pickle mask cache
        state = self.__dict__
        state['_get_masks_for_slice'] = {}
        return state

    def validate_mask_functions(self):
        fns = self.mask_factories
        # 1 MB, magic number L3 cache
        limit = 2**20
        if callable(fns):
            fns = [fns]
        for fn in fns:
            s = len(cloudpickle.dumps(fn))
            if s > limit:
                log.warning(
                    'Mask factory size %s larger than warning limit %s, may be inefficient'
                    % (s, limit)
                )

    def __len__(self):
        if self._length is not None:
            return self._length
        elif not callable(self.mask_factories):
            return len(self.mask_factories)
        else:
            return len(self.computed_masks)

    def get_for_idx(self, scheme: TilingScheme, idx: int, *args, **kwargs):
        slice_ = scheme[idx]
        return self._get(slice_, *args, **kwargs)

    def get_for_sig_slice(self, sig_slice: Slice, *args, **kwargs):
        """
        Same as `get`, but without calling `discard_nav()` on the slice
        """
        return self._get(sig_slice, *args, **kwargs)

    def get(self, key: Slice, dtype=None, sparse_backend=None, transpose=True, backend=None):
        if not isinstance(key, Slice):
            raise TypeError(
                "MaskContainer.get() can only be called with "
                "DataTile/Slice/Partition instances"
            )
        return self._get(key.discard_nav(), dtype, sparse_backend, transpose, backend)

    def _get(self, slice_: Slice, dtype=None, sparse_backend=None, transpose=True, backend=None):
        if backend is None:
            backend = self.backend
        return self.get_masks_for_slice(
            slice_,
            dtype=dtype,
            sparse_backend=sparse_backend,
            transpose=transpose,
            backend=backend
        )

    @property
    def dtype(self):
        if self._dtype is None:
            return self.computed_masks.dtype
        else:
            return self._dtype

    @property
    def use_sparse(self) -> Union[SparseSupportedT, Literal[False]]:
        # As far as possible use_sparse was resolved at __init__
        # but if we don't know if the masks are sparse we may still arrive
        # here with self._use_sparse is None
        if self._use_sparse is None:
            if is_sparse(self.computed_masks):
                # The first time the condition is hit will cause
                # mask computation but on subsequent tries we will
                # fall through to the normal return
                self._use_sparse = self._default_sparse
            else:
                self._use_sparse = False
        return self._use_sparse

    def _compute_masks(self) -> Union[np.ndarray, sparse.COO, sparse.GCXS]:
        """
        Call mask factories and combine into a mask stack

        Uses the internal attr self._use_sparse, which could be None
        if we were unable to resolve the sparse mode at __init__
        If self._use_sparse is None and all masks are sparse then will
        return a sparse stack else return a dense stack
        Otherwise if self._use_sparse is simply False then return
        dense, anything else return as a sparse stack

        Returns
        -------
        an array-like mask stack with contents as they were
        created by the factories
        """
        mask_slices = []
        if callable(self.mask_factories):
            raw_masks = self.mask_factories()
            mask_slices.append(raw_masks)
        else:
            for f in self.mask_factories:
                m = f()
                # Scipy.sparse is always 2D, so we have to convert here
                # before reshaping
                if scipy.sparse.issparse(m):
                    m = sparse.COO.from_scipy_sparse(m)
                # We reshape to be a stack of 1 so that we can unify code below
                m = m.reshape((1, ) + m.shape)
                mask_slices.append(m)

        # Fully resolve _use_sparse based on sparsity of masks.
        # The return type (sparse or dense) from this function
        # is used to resolve _use_sparse permanently in the
        # self.use_sparse property method
        masks_are_sparse = all(is_sparse(m) for m in mask_slices)
        use_sparse = self._use_sparse
        if use_sparse is None:
            if masks_are_sparse:
                use_sparse = self._default_sparse
            else:
                use_sparse = False

        if use_sparse is not False:
            # Conversion to correct back-end will happen later
            # Use sparse.pydata because it implements the array interface
            # which makes mask handling easier
            masks = sparse.concatenate(
                [to_sparse(m) for m in mask_slices]
            )
        else:
            masks = np.concatenate(
                [to_dense(m) for m in mask_slices]
            )
        return masks

    def get_masks_for_slice(self, slice_, dtype=None, sparse_backend=None,
            transpose=True, backend='numpy'):
        if dtype is None:
            dtype = self.dtype
        if sparse_backend is None:
            sparse_backend = self.use_sparse
        if backend is None:
            backend = self.backend
        key = (dtype, sparse_backend, transpose, backend)
        if key not in self._get_masks_for_slice:
            self._get_masks_for_slice[key] = _make_mask_slicer(
                self.computed_masks,
                dtype=dtype,
                sparse_backend=sparse_backend,
                transpose=transpose,
                backend=backend
            )
        return self._get_masks_for_slice[key](slice_)

    @property
    def computed_masks(self):
        if self._computed_masks is None:
            self._computed_masks = self._compute_masks()
        return self._computed_masks



class UDFException(Exception):
    """
    Raised when the UDF interface is somehow misused
    """
    pass


class ExecutorSpecException(Exception):
    """
    Raised when there is an error specifying an
    Executor or its resources / workers
    """
    pass


class UDFRunCancelled(Exception):
    """
    Raised when the UDF run was cancelled, either when the job was cancelled
    using :meth:`AsyncJobExecutor.cancel`, or when the underlying data source
    was interrupted.
    """



import sys
import builtins
import platform

import pytest


if platform.system() != 'Windows':
    pytest.skip("Skipping Windows-specific tests", allow_module_level=True)


real_import = builtins.__import__


def monkey_notfound(name, *args, **kwargs):
    if name in ('win32security', 'pywintypes'):
        raise ModuleNotFoundError(f"Mocked module not found {name}")
    return real_import(name, *args, **kwargs)


def monkey_importerror(name, *args, **kwargs):
    if name in ('win32security', ):
        raise ImportError(f"Mocked import error {name}")
    return real_import(name, *args, **kwargs)


def test_import_broken(monkeypatch):
    monkeypatch.delitem(sys.modules, 'win32security', raising=False)
    monkeypatch.delitem(sys.modules, 'libertem.common.win_tweaks', raising=False)
    monkeypatch.setattr(builtins, '__import__', monkey_importerror)

    # Self test
    with pytest.raises(ImportError):
        import win32security  # NOQA: F401

    from libertem.common.win_tweaks import get_owner_name

    # Make sure the dummy fallback version works
    # This would fail with the actual get_owner_name() function
    assert get_owner_name('/asdf/bla', None) == '<Unknown>'


def test_import_missing(monkeypatch):
    monkeypatch.delitem(sys.modules, 'win32security', raising=False)
    monkeypatch.delitem(sys.modules, 'libertem.common.win_tweaks', raising=False)
    monkeypatch.setattr(builtins, '__import__', monkey_notfound)

    # Self test
    with pytest.raises(ModuleNotFoundError):
        import win32security  # NOQA: F401

    from libertem.common.win_tweaks import get_owner_name

    # Make sure the dummy fallback version works
    # This would fail with the actual get_owner_name() function
    assert get_owner_name('/asdf/bla', None) == '<Unknown>'



import numpy as np
import pathlib
import dask
import itertools
import dask.array as da


def _create_chunk(*, chunk_shape, dtype, value=1, **kwargs):
    return np.ones(chunk_shape, dtype=dtype) * value


def _mmap_load_chunk(*, dataset_shape, dtype, filename, sl, offset=0, **kwargs):
    data = np.memmap(filename, mode='r', shape=dataset_shape, dtype=dtype, offset=offset)
    return data[sl]


def _take_n(iterable, n):
    """
    Split a interable into chunks of length n with the final element
    being the remainder len < n if n does not divide evenly
    """
    len_iter = len(iterable)
    return [iterable[i: min(i + n, len_iter)] for i in range(0, len_iter, n)]


def _reshape_list(to_reshape, new_shape):
    """
    Recursively reshape a list into new_shape following C-ordering
    Unlike numpy.reshape this will happily create a ragged array when
    the dimensions in new_shape do not divide perfectly into to_reshape!
    """
    left = new_shape[:-1]
    n = new_shape[-1]
    reshaped = _take_n(to_reshape, n)
    if len(left) == 0:
        return reshaped[0]
    else:
        return _reshape_list(reshaped, left)


def _blocksize_to_dim_slices(blocksize, dim):
    """
    Generate the slices along a dimension of length dim
    to split dim into chunks of length blocksize, with the final chunk
    being the remainder if dim % blocksize != 0

    A blocksize of -1 is converted to a None slice
    """
    slices = []
    if blocksize == -1:
        slices.append(slice(None))
    else:
        for index in range(0, dim, blocksize):
            chunk_size = min(blocksize, dim - index)
            slices.append(slice(index, index + chunk_size))
    return slices


def _slices_to_chunk_shape(slices, full_shape):
    """
    Generate the shape tuple of the array resulting from applying
    the sequence of slices to an array with shape full_shape
    """
    chunk_shape = []
    for s, dim in zip(slices, full_shape):
        if s.start is not None and s.stop is not None:
            chunk_shape.append(s.stop - s.start)
        elif s.start is None and s.stop is None:
            chunk_shape.append(dim)
        else:
            raise NotImplementedError(f'Open-ended slices are currently unhandled {s}')
    return tuple(chunk_shape)


def _get_block_slices(blocksizes, shape):
    """
    Converts blocksizes into a list of list of slices along each dimension in shape
    An integer blocksizes is treated as a blocksize for the 0th dimension only
    A blocksize of -1 is treated as slice(None) for that dimension
    """
    if isinstance(blocksizes, int):
        # assume chunking only in first dimension
        blocksizes = (blocksizes,) + ((-1,) * len(shape[1:]))
    assert len(blocksizes) == len(shape), ('Must supply a blocksize for every ',
                                           'dimension, (-1 == no chunking), or int')
    return [_blocksize_to_dim_slices(bsize, dim) for bsize, dim in zip(blocksizes, shape)]


def _mk_dask_from_delayed(shape, chunking, dtype='float32', filename=None, value=None):
    """
    Create a dask array by combining individually created blocks

    If filename is not None will load from file using np.memmap
    otherwise will generate numbered partitions using np.ones * chunk_idx
    or partitions of uniform value if value is not None
    """
    if filename is not None:
        create = dask.delayed(_mmap_load_chunk, name='create_chunk', pure=True, traverse=False)
        filename = pathlib.Path(filename)
    else:
        create = dask.delayed(_create_chunk, name='create_chunk', pure=True, traverse=False)

    slices_per_dim = _get_block_slices(chunking, shape)
    blocks = []
    # rightmost advances fastest with itertools.product
    for chunk_idx, chunk_slices in enumerate(itertools.product(*slices_per_dim)):
        chunk_value = chunk_idx if value is None else value
        chunk_shape = _slices_to_chunk_shape(chunk_slices, shape)
        chunk = dask.array.from_delayed(
            create(
                dataset_shape=shape,
                chunk_shape=chunk_shape,
                dtype=dtype,
                value=chunk_value,
                filename=filename,
                sl=chunk_slices
            ),
            shape=chunk_shape,
            dtype=dtype
        )
        blocks.append(chunk)

    nblocks_per_dim = tuple(len(ss) for ss in slices_per_dim)
    blocks = _reshape_list(blocks, nblocks_per_dim)
    return da.block(blocks)



import numpy as np

from libertem.udf.raw import PickUDF
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


def test_pick(lt_ctx, delayed_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    # data = np.ones((16, 16, 16, 16), dtype="float32")
    # data = np.arange(0, 16*16*16*16, dtype="float32").reshape((16, 16, 16, 16))
    dataset = MemoryDataSet(data=data, tileshape=(3, 7, 16),
                            num_partitions=7, sig_dims=2)
    roi = np.random.choice([True, False], size=dataset.shape.nav)
    roi[0] = True

    udf = PickUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=udf, roi=roi)
    res_delayed = delayed_ctx.run_udf(dataset=dataset, udf=udf, roi=roi)

    assert np.allclose(data[roi], res['intensity'].data)
    assert np.allclose(data[roi], res_delayed['intensity'].data)

    assert data.dtype == res['intensity'].data.dtype
    assert data.dtype == res_delayed['intensity'].data.dtype


def test_pick_empty_roi(lt_ctx, delayed_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(3, 7, 7),
                            num_partitions=7, sig_dims=2)
    roi = np.zeros(dataset.shape.nav, dtype=bool)

    udf = PickUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=udf, roi=roi)
    res_delayed = delayed_ctx.run_udf(dataset=dataset, udf=udf, roi=roi)

    assert np.allclose(data[roi], res['intensity'].data)
    assert np.allclose(data[roi], res_delayed['intensity'].data)

    assert data[roi].shape == res['intensity'].data.shape
    assert data[roi].shape == res_delayed['intensity'].data.shape

    assert data.dtype == res['intensity'].data.dtype
    assert data.dtype == res_delayed['intensity'].data.dtype



import asyncio
import uuid
import pytest
from unittest import mock

from libertem.executor.base import AsyncAdapter
from libertem.executor.dask import DaskJobExecutor
from libertem.web.event_bus import EventBus
from libertem.web.state import JobState, ExecutorState, SharedState


pytestmark = []


@pytest.mark.asyncio
async def test_job_state_remove(async_executor):
    event_bus = EventBus()
    executor_state = ExecutorState(event_bus=event_bus)
    try:
        await executor_state.set_executor(async_executor, {})
        job_state = JobState(executor_state=executor_state)

        job_id = str(uuid.uuid4())
        analysis_id = str(uuid.uuid4())
        dataset_id = str(uuid.uuid4())

        job_state.register(job_id, analysis_id, dataset_id)

        assert job_state[job_id] == {
            "id": job_id,
            "analysis": analysis_id,
            "dataset": dataset_id,
        }
        assert job_state.get_for_dataset_id(dataset_id) == {job_id}
        assert job_state.get_for_analysis_id(analysis_id) == {job_id}

        await job_state.remove(job_id)

        print(job_state.jobs)

        with pytest.raises(KeyError):
            job_state[job_id]

        assert job_state.get_for_dataset_id(dataset_id) == set()
        assert job_state.get_for_analysis_id(analysis_id) == set()
        assert len(job_state.jobs) == 0
    finally:
        executor_state.shutdown()


@pytest.mark.web_api
@pytest.mark.asyncio
async def test_preload_executor(tmpdir_factory):
    workdir = tmpdir_factory.mktemp('preload_workdir')
    event_bus = EventBus()
    executor_state = ExecutorState(event_bus=event_bus)
    sync_exec = None
    try:
        state = SharedState(executor_state=executor_state)
        executor_state.set_local_directory(workdir)
        executor_state.set_preload(())

        import tornado.ioloop
        io_loop = tornado.ioloop.IOLoop.current()
        assert io_loop is not None

        await state.create_and_set_executor(
            {
                'cpus': 2,
                'cudas': 0,
            },
        )
        assert isinstance(state.executor_state.executor, AsyncAdapter)
        sync_exec = state.executor_state.executor.ensure_sync()
        assert isinstance(sync_exec, DaskJobExecutor)
        assert sync_exec.is_local
        resources = sync_exec.get_resource_details()
        assert resources[0]['cpu'] == 2
        assert resources[0]['cuda'] == 0
        assert resources[0]['service'] == 1

        def test_fn():
            return 42

        assert 42 == sync_exec.run_function(test_fn)
    finally:
        if sync_exec is not None:
            sync_exec.close()
        executor_state.shutdown()


@pytest.mark.asyncio
async def test_snooze_last_activity(async_executor):
    event_bus = EventBus()
    executor_state = ExecutorState(event_bus=event_bus)
    await executor_state.set_executor(async_executor, {})

    try:
        # each of these activities should reset the last activity timer:
        executor_state._update_last_activity = mock.Mock()
        _ = await executor_state.get_executor()
        executor_state._update_last_activity.assert_called()

        executor_state._update_last_activity = mock.Mock()
        _ = await executor_state.get_context()
        executor_state._update_last_activity.assert_called()

        executor_state._update_last_activity = mock.Mock()
        _ = executor_state.get_cluster_params()
        executor_state._update_last_activity.assert_called()

        # the previous executor is closed in `set_executor`, which we don't want to
        # happen here:
        executor_state.executor = None
        executor_state._update_last_activity = mock.Mock()
        await executor_state.set_executor(async_executor, {})
        executor_state._update_last_activity.assert_called()
    finally:
        executor_state.shutdown()


@pytest.mark.asyncio
async def test_snooze_explicit(local_cluster_url):
    """
    We can snooze and unsnooze the executor that is part of the ExecutorState:
    """
    event_bus = EventBus()
    executor_state = ExecutorState(event_bus=event_bus)
    pool = AsyncAdapter.make_pool()
    try:
        params = {
            "connection": {
                "type": "tcp",
                "address": local_cluster_url,
            }
        }
        executor = await executor_state.make_executor(params, pool)
        await executor_state.set_executor(executor, params)

        await executor_state.snooze()
        assert executor_state._is_snoozing
        assert executor_state.executor is None
        assert executor_state.context is None

        await executor_state.unsnooze()
        assert not executor_state._is_snoozing
        assert executor_state.executor is not None
        assert executor_state.context is not None
        # these two work without raising an exception:
        await executor_state.get_executor()
        await executor_state.get_context()
    finally:
        pool.shutdown()
        executor_state.shutdown()


@pytest.mark.asyncio
async def test_snooze_explicit_keep_alive(local_cluster_url):
    """
    We can't snooze if keep-alive is nonzero
    """
    event_bus = EventBus()
    executor_state = ExecutorState(event_bus=event_bus)
    pool = AsyncAdapter.make_pool()
    try:
        params = {
            "connection": {
                "type": "tcp",
                "address": local_cluster_url,
            }
        }
        executor = await executor_state.make_executor(params, pool)
        await executor_state.set_executor(executor, params)

        # if we are in at least one keep-alive section, we can't snooze:
        with executor_state.keep_alive():
            assert executor_state._keep_alive > 0
            await executor_state.snooze()
            assert not executor_state._is_snoozing
            assert executor_state.executor is not None
            assert executor_state.context is not None

        # keep-alive can nest:
        with executor_state.keep_alive():
            with executor_state.keep_alive():
                assert executor_state._keep_alive > 0
                await executor_state.snooze()
                assert not executor_state._is_snoozing
                assert executor_state.executor is not None
                assert executor_state.context is not None

        # afterwards, we can snooze again:
        assert executor_state._keep_alive == 0
        await executor_state.snooze()
        assert executor_state._is_snoozing
        assert executor_state.executor is None
        assert executor_state.context is None

        await executor_state.unsnooze()
        assert not executor_state._is_snoozing
        assert executor_state.executor is not None
        assert executor_state.context is not None
        # these two work without raising an exception:
        await executor_state.get_executor()
        await executor_state.get_context()
    finally:
        pool.shutdown()
        executor_state.shutdown()


@pytest.mark.asyncio
async def test_snooze_by_activity(local_cluster_url):
    """
    Test that timer-based snoozing works
    """
    pool = AsyncAdapter.make_pool()
    event_bus = EventBus()
    executor_state = ExecutorState(snooze_timeout=0.01, event_bus=event_bus)
    try:
        params = {
            "connection": {
                "type": "tcp",
                "address": local_cluster_url,
            }
        }
        executor = await executor_state.make_executor(params, pool)
        # we must check very frequently; by default we only check twice a minute
        # to keep activity low:
        executor_state._snooze_check_interval = 0.01
        await executor_state.set_executor(executor, params)

        await asyncio.sleep(0.1)
        # after this sleep, the executor should be snoozed, as we had about ten
        # opportunities to snooze in between:
        assert executor_state._is_snoozing
        assert executor_state.executor is None
        assert executor_state.context is None

        # and this should directly unsnooze the executor
        # (we need to change the timeout etc. here, before we trigger the unsnooze,
        # to make sure we don't directly snooze again):
        executor_state._snooze_timeout = 3600.0
        _ = await executor_state.get_executor()
        assert not executor_state._is_snoozing
        assert executor_state.executor is not None
        assert executor_state.context is not None

        # should not raise an exception, there should be a snooze message
        # in here (raises Empty otherwise):
        event_bus.get()
    finally:
        executor_state.shutdown()



import json
import pytest
import websockets
import nbformat as nbf

from utils import assert_msg
from nbconvert.preprocessors import ExecutePreprocessor

from aio_utils import (
    create_connection, consume_task_results, create_default_dataset, create_analysis,
    create_job_for_analysis, create_update_compound_analysis,
)
from libertem.io.writers.results import formats  # NOQA: F401


pytestmark = [pytest.mark.web_api]


@pytest.mark.asyncio
async def test_copy_notebook(
    default_raw, base_url, tmpdir_factory, http_client, server_port, local_cluster_url,
    default_token,
):
    datadir = tmpdir_factory.mktemp('test_copy')

    await create_connection(base_url, http_client, local_cluster_url, default_token)

    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "APPLY_RING_MASK",
                "analyses": [],
            }, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, details={
                "analysisType": "APPLY_RING_MASK",
                "parameters": {
                    "shape": "ring",
                    "cx": 8,
                    "cy": 8,
                    "ri": 5,
                    "ro": 8,
                }
            }, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "APPLY_RING_MASK",
                "analyses": [analysis_uuid],
            }, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)
        download_url = "{}/api/compoundAnalyses/{}/copy/notebook/?token={}".format(
            base_url, ca_uuid, default_token,
        )
        async with http_client.get(download_url) as resp:
            assert resp.status == 200
            code = await resp.json()
            assert 'dependency' in code
            assert 'initial_setup' in code
            assert 'ctx' in code
            assert 'dataset' in code
            assert 'analysis' in code['analysis'][0]
            assert 'plot' in code['analysis'][0]

        # we are done with this analysis, clean up:
        # this will also remove the associated jobs
        async with http_client.delete(analysis_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'ANALYSIS_REMOVED')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')

        code = "\n".join([
                    code['dependency'],
                    code['initial_setup'],
                    code['ctx'],
                    code['dataset'],
                    code['analysis'][0]['analysis'],
                    code['analysis'][0]['plot'][0]
                ])
        nb = nbf.v4.new_notebook()
        cell = nbf.v4.new_code_cell(code)
        nb['cells'].append(cell)
        ep = ExecutePreprocessor(timeout=600)
        ep.preprocess(nb, {"metadata": {"path": datadir}})



import io
import os

import numpy as np
import nbformat
from temp_utils import _get_hdf5_params
from libertem.web.notebook_generator.notebook_generator import notebook_generator
from nbconvert.preprocessors import ExecutePreprocessor
import pytest


@pytest.mark.slow
def test_com_default(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)

    analysis = [{
        "analysisType": 'CENTER_OF_MASS',
        "parameters": {
            'shape': 'com',
            'cx': 0,
            'cy': 0,
            'r': 8,
            'ri': 4,
        }
    }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=30)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    channels = [
        "field",
        "magnitude",
        "divergence",
        "curl",
        "x",
        "y"
    ]
    results = {}
    for channel in channels:
        data_path = os.path.join(datadir, f"com_result_{channel}.npy")
        results[channel] = np.load(data_path)

    com_analysis = lt_ctx.create_com_analysis(
        dataset=hdf5_ds_2,
        cx=0,
        cy=0,
        mask_radius=8,
        mask_radius_inner=4,
    )
    expected = lt_ctx.run(com_analysis)

    assert np.allclose(results["field"], expected["field"].raw_data)
    assert np.allclose(results["magnitude"], expected["magnitude"].raw_data)
    assert np.allclose(results["divergence"], expected["divergence"].raw_data)
    assert np.allclose(results["curl"], expected["curl"].raw_data)
    assert np.allclose(results["x"], expected["x"].raw_data)
    assert np.allclose(results["y"], expected["y"].raw_data)


@pytest.mark.slow
def test_com_rotflip(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)

    analysis = [{
            "analysisType": 'CENTER_OF_MASS',
            "parameters": {
                        'shape': 'com',
                        'cx': 0,
                        'cy': 0,
                        'r': 8,
                        'flip_y': True,
                        'scan_rotation': -42.23
                        }
    }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    channels = [
            "field",
            "magnitude",
            "divergence",
            "curl",
            "x",
            "y"
    ]
    results = {}
    for channel in channels:
        data_path = os.path.join(datadir, f"com_result_{channel}.npy")
        results[channel] = np.load(data_path)

    com_analysis = lt_ctx.create_com_analysis(
                                dataset=hdf5_ds_2,
                                cx=0,
                                cy=0,
                                mask_radius=8,
                                flip_y=True,
                                scan_rotation=-42.23,
                            )
    expected = lt_ctx.run(com_analysis)

    assert np.allclose(results["field"], expected["field"].raw_data)
    assert np.allclose(results["magnitude"], expected["magnitude"].raw_data)
    assert np.allclose(results["divergence"], expected["divergence"].raw_data)
    assert np.allclose(results["curl"], expected["curl"].raw_data)
    assert np.allclose(results["x"], expected["x"].raw_data)
    assert np.allclose(results["y"], expected["y"].raw_data)



from unittest import mock


import pytest


from libertem.io.dataset.cached import CacheStats, CacheItem, OrphanItem


@pytest.fixture
def cs():
    cs = CacheStats(":memory:")
    cs.initialize_schema()
    yield cs
    cs.close()


@pytest.fixture
def ci():
    return CacheItem(
        dataset="deadbeef",
        partition=5,
        size=768,
        path="/tmp/dont_care",
    )


@pytest.fixture
def oi():
    return OrphanItem(
        size=512,
        path="/tmp/dont_care",
    )


def test_cache_stats_starts_connected(cs):
    assert cs._conn is not None


def test_initialize_schema_idempotent(cs):
    cs.initialize_schema()  # already called in fixture, should not raise here


def test_first_miss(cs, ci):
    with mock.patch('time.time', side_effect=lambda: 42):
        cs.record_miss(ci)
    ds_stats = cs.get_stats_for_dataset("deadbeef")
    stats_items = list(ds_stats.items())
    assert len(stats_items) == 1
    si = stats_items[0]
    assert si[0] == 5
    assert si[1]["last_access"] == 42
    assert si[1]["hits"] == 0


def test_first_hits(cs, ci):
    # Unfortunately, someone is calling time.time() under the hood, and doing it
    # in a nondeterministic way. What I would have liked to test here: passing
    # side_effect=[21, 42] and making sure the 21 is overwritten by the 42.
    # Now we don't know if the first or second value survives.
    with mock.patch('time.time', side_effect=lambda: 42):
        cs.record_miss(ci)
        cs.record_hit(ci)
    ds_stats = cs.get_stats_for_dataset("deadbeef")
    stats_items = list(ds_stats.items())
    assert len(stats_items) == 1
    si = stats_items[0]
    assert si[0] == 5
    assert si[1]["last_access"] == 42
    assert si[1]["hits"] == 1


def test_eviction(cs, ci):
    with mock.patch('time.time', side_effect=lambda: 42):
        cs.record_miss(ci)
        cs.record_hit(ci)
        cs.record_eviction(ci)
    ds_stats = cs.get_stats_for_dataset("deadbeef")
    stats_items = list(ds_stats.items())
    assert len(stats_items) == 0


def test_record_orphan(cs, oi):
    cs.maybe_orphan(oi)

    cursor = cs.query("SELECT * FROM stats")
    assert len(cursor.fetchall()) == 0

    orphans = cs.get_orphans()
    assert len(orphans) == 1


def test_hit_after_orphan(cs, ci, oi):
    cs.maybe_orphan(oi)

    orphans = cs.get_orphans()
    assert len(orphans) == 1

    with mock.patch('time.time', side_effect=lambda: 42):
        cs.record_hit(ci)
    ds_stats = cs.get_stats_for_dataset("deadbeef")
    stats_items = list(ds_stats.items())
    assert len(stats_items) == 1
    si = stats_items[0]
    assert si[0] == 5
    assert si[1]["last_access"] == 42
    assert si[1]["hits"] == 1

    orphans = cs.get_orphans()
    assert len(orphans) == 0


def test_miss_after_orphan(cs, ci, oi):
    """
    db/fs out of sync: the orphan was deleted from fs, then miss'd
    """
    cs.maybe_orphan(oi)

    orphans = cs.get_orphans()
    assert len(orphans) == 1

    with mock.patch('time.time', side_effect=lambda: 42):
        cs.record_miss(ci)
    ds_stats = cs.get_stats_for_dataset("deadbeef")
    stats_items = list(ds_stats.items())
    assert len(stats_items) == 1
    si = stats_items[0]
    assert si[0] == 5
    assert si[1]["last_access"] == 42
    assert si[1]["hits"] == 0

    orphans = cs.get_orphans()
    assert len(orphans) == 0



import logging
import warnings
warnings.warn(
    'gridmatching and fullmatch are moved to '
    'libertem_blobfinder.common in libertem-blobfinder>=0.6 '
    'and are slated for removal in LiberTEM 0.16.',
    DeprecationWarning
)

import numpy as np  # NOQA: 402
import hdbscan  # NOQA: 402

import libertem.analysis.gridmatching as grm  # NOQA: 402
from libertem.utils import make_polar, make_cartesian  # NOQA: 402


log = logging.getLogger(__name__)


class NotFoundException(Exception):
    pass


class FullMatcher(grm.Matcher):
    '''
    Extension of :class:`~libertem.analysis.gridmatching.Matcher` will full matching

    Include the ability to guess grid parameters from a point cloud. This is separated
    from the other code since it currently only works with :class:`~hdbscan.HDBSCAN`,
    which can be problematic
    to install on some platforms. For that reason it is an optional dependency.

    Parameters
    ----------

    tolerance : float
        Position tolerance in px for peaks to be considered matches
    min_weight : float
        Minimum peak elevation of a peak to be considered for matching
    min_match : int
        Minimum number of matching peaks to be considered a match.
    min_angle : float
        Minimum angle in radians between two vectors to be considered candidates
    min_points : int
        Minimum points to try clustering matching. Otherwise match directly
    min_delta : float
        Minimum length of a potential grid vector
    max_delta : float
        Maximum length of a potential grid vector
    min_candidates : int
        Minimum number of candidates to consider clustering matching successful.
        If not enough are found, the algorithm uses a brute-force search with all
        pairwise vectors between points
    max_candidates : int
        Maximum number of candidates to return from clustering matching
    clusterer
        Instance of sklearn.cluster compatible clusterer. Default is :class:`~hdbscan.HDBSCAN`.
    min_cluster_size_fraction : float
        Tuning parameter for clustering matching with :class:`~hdbscan.HDBSCAN`.
        Larger values allow
        smaller or fuzzier clusters. This is used to adapt the :code:`min_cluster_size`
        parameter of :class:`~hdbscan.HDBSCAN` dynamically to the number of points to be
        matched.
        Set this to :code:`None` to disable dynamic adjustment of :code:`min_cluster_size`.
        If you like to set :code:`min_cluster_size` to a constant value, you can
        set this to :code:`None` and additionally set the :code:`clusterer` parameter with
        your own clusterer object to have direct control over all parameters.
    min_samples_fraction : float
        Tuning parameter for clustering matching with :class:`~hdbscan.HDBSCAN`.
        Larger values allow
        smaller or fuzzier clusters. This is used to adapt the :code:`min_samples`
        parameter of :class:`~hdbscan.HDBSCAN` dynamically to the number of points to be
        matched.
        Set this to :code:`None` to disable dynamic adjustment of :code:`min_samples`.
        If you like to set :code:`min_samples` to a constant value, you can
        set this to :code:`None` and additionally set the :code:`clusterer` parameter with
        your own clusterer object to have direct control over all parameters.
    '''
    def __init__(
            self, tolerance=3, min_weight=0.1, min_match=3, min_angle=np.pi/10,
            min_points=10, min_delta=0, max_delta=np.inf, min_candidates=3,
            max_candidates=7, clusterer=None, min_cluster_size_fraction=4,
            min_samples_fraction=20):

        super().__init__(tolerance=tolerance, min_weight=min_weight, min_match=min_match)
        if clusterer is None:
            clusterer = hdbscan.HDBSCAN()
        self.min_angle = min_angle
        self.min_points = min_points
        self.min_delta = min_delta
        self.max_delta = max_delta
        self.min_candidates = min_candidates
        self.max_candidates = max_candidates
        self.clusterer = clusterer
        self.min_cluster_size_fraction = min_cluster_size_fraction
        self.min_samples_fraction = min_samples_fraction

    def full_match(
            self, centers, zero=None, cand=None,
            refineds=None, peak_values=None, peak_elevations=None):
        '''
        This function extracts a list of Match objects as well two PointSelection objects
        for unmatched and weak points from correlation_result and zero point.
        The zero point is included in each of the matches because it is shared between all grids.

        Parameters
        ----------
        centers : numpy.ndarray
            numpy.ndarray of shape (n, 2) with integer centers (y, x) of peaks. This would typically
            be extracted with :meth:`libertem_blobfinder.common.correlation.get_peaks`
        zero : numpy.ndarray
            Zero point as numpy array (y, x).
        cand : list or numpy.ndarray
            Optional list of candidate vectors (y, x) to use in a first matching round before
            guessing.
        refineds : numpy.ndarray
            numpy.ndarray of shape (n, 2) with float centers (y, x) of peaks (subpixel refinement)
        peak_values : numpy.ndarray
            numpy.ndarray of shape (n,) with float maxima of correlation map of peaks
        peak_elevations : numpy.ndarray
            numpy.ndarray of shape (n,) with float elevation of correlation map of peaks.
            See :meth:`libertem_blobfinder.base.correlation.peak_elevation` for details.

        Returns
        -------
        Tuple[List[libertem.analysis.gridmatching.Match, ...],\
        libertem.analysis.gridmatching.PointSelection,\
        libertem.analysis.gridmatching.PointSelection]
            matches: list of :class:`~libertem.analysis.gridmatching.Match` instances,

            unmatched: instance of :class:`~libertem.analysis.gridmatching.PointSelection`,

            weak: instance of :class:`~libertem.analysis.gridmatching.PointSelection`

        Example
        -------

        >>> peaks = np.array([
        ...     # First peak is zero if not specified otherwise
        ...     # Base lattice vectors (32, 0) and (0, 32)
        ...     (64, 64),
        ...     (32, 32), (32, 64), (32, 96),
        ...     (64, 32), (64, 96),
        ...     (96, 32), (96, 64), (96, 96),
        ... ])
        >>> matcher = FullMatcher()
        >>> (matches, unmatched, weak) = matcher.full_match(peaks)
        >>> m = matches[0]
        >>> assert np.allclose(m.zero, (64, 64))
        >>> assert np.allclose(m.a, (32, 0))
        >>> assert np.allclose(m.b, (0, 32))
        '''
        class ExitException(Exception):
            pass

        if zero is None:
            zero = centers[0]

        corr = grm.CorrelationResult(
            centers=centers,
            refineds=refineds,
            peak_values=peak_values,
            peak_elevations=peak_elevations,
        )

        matches = []

        filt = corr.peak_elevations >= self.min_weight

        working_set = grm.PointSelection(corr, selector=filt)

        zero_selector = np.array([
            np.allclose(corr.centers[i], zero)
            + np.allclose(corr.refineds[i], zero)
            for i in range(len(corr))
        ], dtype=bool)

        def listed(working_set, polar_cand):
            return polar_cand

        def guess(working_set, polar_cand):
            return self._candidates(working_set.refineds)

        if cand is not None:
            polar_cand = size_filter(
                make_polar(np.array(cand)),
                min_delta=self.min_delta,
                max_delta=self.max_delta
            )
            candidate_methods = [listed, guess]
        else:
            polar_cand = None
            candidate_methods = [guess]

        while True:
            new_selector = np.copy(working_set.selector)
            # First, find good candidate
            # Expensive operation, should be done on smaller sample
            # or sum frame result, at least for first passes to match majority
            # of peaks
            polar_candidate_vectors = candidate_methods[0](working_set, polar_cand)

            match = self._find_best_vector_match(
                point_selection=working_set, zero=zero,
                candidates=polar_candidate_vectors)
            if match is None:
                candidate_methods = candidate_methods[1:]
                if len(candidate_methods) == 0:
                    break
                else:
                    continue
            matches.append(match)
            # remove the ones that have been matched
            new_selector[match.selector] = False
            if np.count_nonzero(new_selector) >= self.min_match:
                # Add zero point that is shared by all patterns
                new_selector[zero_selector] = True
                working_set = working_set.derive(selector=new_selector)
            else:
                break
        if matches:
            new_selector[zero_selector] = False
        unmatched = working_set.derive(selector=new_selector)
        weak = grm.PointSelection(corr, selector=np.logical_not(filt))
        return (matches, unmatched, weak)

    def make_polar_vectors(self, coords):
        '''
        Calculate all unique pairwise connecting polar vectors between points in coords.

        The pairwise connecting vectors are converted to polar coordinates and
        filtered with parameters :py:attr:`~min_delta` and :py:attr:`~max_delta`
        to avoid calculating for unwanted higher order or random smaller vectors.

        All calculated vectors have a positive or zero x direction.
        '''
        # sort by x coordinate so that we have always positive x difference vectors
        sort_indices = np.argsort(coords[:, 1])
        coords = coords[sort_indices]
        i, j = np.mgrid[0: len(coords), 0: len(coords)]
        selector = j > i
        deltas = coords[j[selector]] - coords[i[selector]]
        polar = make_polar(deltas)
        return size_filter(polar, self.min_delta, self.max_delta)

    def check(self, match):
        if len(match) < self.min_match:
            return False
        papb = make_polar(np.array([match.a, match.b]))
        if len(size_filter(papb, self.min_delta, self.max_delta)) != 2:
            return False
        return angle_check(papb[0:1], papb[1:2], self.min_angle)

    def _tumble(self, point_selection, match):
        if not self.check(match):
            return None
        match = match.weighted_optimize()
        if not self.check(match):
            return None
        match = self._match_all(
            point_selection=point_selection, zero=match.zero, a=match.a, b=match.b)
        if not self.check(match):
            return None
        match = match.weighted_optimize()
        if not self.check(match):
            return None
        else:
            return match

    def _do_match(self, point_selection: grm.PointSelection, zero, polar_vectors):
        '''
        Return a list with matches of all pairwise combinations of polar_vectors
        '''
        match_list = []
        # we test all pairs of candidate vectors
        # and populate match_matrix
        for i in range(len(polar_vectors)):
            for j in range(i + 1, len(polar_vectors)):
                a = polar_vectors[i]
                b = polar_vectors[j]
                # too parallel, not good lattice vectors
                if not angle_check(np.array([a]), np.array([b]), self.min_angle):
                    continue

                if a[0] > b[0]:
                    bb = a
                    aa = b
                else:
                    aa = a
                    bb = b
                aa, bb = make_cartesian(np.array([aa, bb]))
                try:
                    match = self._match_all(
                        point_selection=point_selection, zero=zero, a=aa, b=bb)
                    match = self._tumble(point_selection, match)
                except np.linalg.LinAlgError:
                    continue
                if match is not None:
                    match_list.append(match)
        return match_list

    def _find_best_vector_match(self, point_selection: grm.PointSelection, zero, candidates):
        '''
        Return the match that matches with the best figure of merit

        Good properties for vectors are
        * Matching many points in the result
        * Orthogonal
        * Equal length
        * Short

        The function implements a heuristic to calculate a figure of merit that boosts
        candidates for each of the criteria that they fulfill or nearly fulfill.

        FIXME improve this on more real-world examples; define test cases.

        FIXME The figure of merit function (fom) could be a parameter, implement if need arises.
        '''
        def fom(m):
            na = np.linalg.norm(m.a)
            nb = np.linalg.norm(m.b)
            # Matching many high-quality points is good
            res = np.sum(m.peak_elevations)**2

            # favor orthogonality
            res *= (np.abs(np.cross(m.a, m.b)) / (na * nb))

            # favor equal length
            res *= ((na * nb) / (na**2 + nb**2))

            return res

        match_list = self._do_match(point_selection, zero, candidates)
        if match_list:
            return max(match_list, key=fom)
        else:
            return None

    def _make_hdbscan_config(self, points):
        # This is handled here because the defaults depend on the number of points
        defaults = {}
        if self.min_cluster_size_fraction is not None:
            defaults['min_cluster_size'] = max(len(points) // self.min_cluster_size_fraction, 2)
        if self.min_samples_fraction is not None:
            defaults['min_samples'] = max(len(points) // self.min_samples_fraction, 1)
        return defaults

    def _hdbscan_candidates(self, points):
        '''
        Use hdbscan clustering to find potential candidates for lattice vectors.

        We rely on the clusterer and its settings to give us tight and well-populated clusters.
        Then we calculate a weighted mean for each cluster.
        In the end we return the shortest matches.
        '''
        # We have special tuning parameters for the default :class:`~hdbscan.HDBSCAN`
        if isinstance(self.clusterer, hdbscan.HDBSCAN):
            defaults = self._make_hdbscan_config(points)
            for key, value in defaults.items():
                setattr(self.clusterer, key, value)
        vectors = self.make_polar_vectors(points)
        self.clusterer.fit(vectors)
        labels = self.clusterer.labels_
        cand = []
        for cluster in range(max(labels) + 1):
            selector = labels == cluster
            v = vectors[selector]
            weights = self.clusterer.probabilities_[selector]
            std = v.std(axis=0)
            mean = np.average(v, axis=0, weights=weights)
            fom = np.linalg.norm(std)
            if fom > self.tolerance:
                # print("too fuzzy")
                continue
            cand.append(mean)
        # return the shortest candidate vectors
        return np.array(sorted(cand, key=lambda d: d[0])[:self.max_candidates])

    def _candidates(self, points):
        polar_vectors = []
        # Enough "flesh" to cluster
        if len(points) > self.min_points:
            # Get some candidates
            polar_vectors = self._hdbscan_candidates(points)
        # Not enough candidates found, use all pairwise vectors as candidates
        # Tighter vector limits mean less candidates from clustering
        # Adjust as needed because the full match is slow for too many points
        if len(polar_vectors) < self.min_candidates:
            if len(points) > self.min_points:
                log.warn(
                    "Matching many points directly, might be computationally intensive: %s" %
                    len(points)
                )
            polar_vectors = self.make_polar_vectors(points)
        return polar_vectors


def size_filter(polar, min_delta, max_delta):
    '''
    Accept a list of polar vectors
    Return a list of polar vectors with length between min_delta and max_delta
    '''
    select = (polar[:, 0] >= min_delta) * (polar[:, 0] <= max_delta)
    return polar[select]


def angle_check(p1, p2, limit):
    '''
    Check if p1 and p2 have an angle difference of at least limit,
    both parallel or antiparallel
    '''
    diff = np.absolute(p1[:, 1] - p2[:, 1]) % np.pi
    return (diff > limit) * (diff < (np.pi - limit))
